/* Faktorisierungsprogramm
   written 11/1998 by Thorsten Reinecke
   
   Fr den mathematischen Hintergrund siehe:
   Riesel, Hans: "Prime Numbers and computer methods for factorization",
   Boston; Basel; Stuttgart: Birkhuser, 1985
    
   Das Programm setzt einige der dort beschriebenen Algorithmen (in freier
   Variation) um.
   Zunchst findet eine Probedivision fr kleine Primzahlen statt.
   Anschlieend wird mit der Pollard-rho-Methode nach weiteren kleineren
   Faktoren gesucht.
   Falls um eins verminderte Faktoren ausschlielich in kleinere Teiler
   zerfallen, so werden diese mittels der Pollard-Phi-Methode entdeckt.
   
   Wenn die bis dahin verbleibende Zahl nachweislich zusammengesetzt
   (also nicht pseudoprim) ist, wird eine Faktorisierung durch
   das Quadratische Sieb versucht.

   Das lineare Gleichungssystem wird dabei simulatan mit den eingefgten
   Relationen gelst.

   In seiner jetzigen Fassung sollte man mit dem Programm
   Zahlen bis um 80 Dezimalstellen in ertrglicher Zeit faktorisieren knnen.
   
   Es fehlt allerdings noch eine ganze Reihe an Optimierungen und Algorithmen.
   So fehlt beispielsweise eine Implementierung eliptischer Kurven zur
   Ermittlung von Faktoren "gemigter Gre". Dies wre das einzige
   Verfahren, mit dem sich momentan auch bei sehr groen Zahlen nichttriviale
   Faktoren entdecken lassen (da seine Laufzeit von der Gre des entdeckten
   Faktors abhngig ist).
   Fr ein effizienteres Sieben andererseits wre vielleicht noch das
   "Number-Field-Sieve" sinnvoll. (Siebverfahren faktorisieren grundstzlich
   in Abhngigkeit von der Gre der zu faktorisierenden Zahl; sie sind
   besonders effizient, wenn die Zahl lediglich zwei Primteiler hat und
   diese von der Stelligkeit her nicht allzuweit voneinander entfernt liegen.)

   Auerdem wre dann das komplette Programm anzupassen und zu optimieren,
   insbesondere auch die lineare Algebra, die man vielleicht auf den
   Hintergrundspeicher auslagern knnte.
   (Auch so ist das Programm schon hufiger verndert worden, so da
   ich es nochmal "sauberschreiben" mte...)
   Mal sehen, wann ich Zeit und Lust dazu habe... ;)
*/

#ifdef USE_OLD_GCC
/* gcc 2.7.x bentigt Fibonacci-Heap, da priority-queues dort anders
   definiert sind... */
#define USE_FIBHEAP
#endif 

#ifndef USE_NETWORK
#ifdef IS_SERVER
#error "SERVER und kein USE_NETWORK"
#endif
#else
#ifndef IS_STANDALONE
const int server_port = 12364;
#include <unix_buffer.H>
#endif
#endif

#ifndef IS_SERVER
#ifndef IS_CLIENT
#undef IS_STANDALONE
#define IS_STANDALONE
#endif
#endif

#ifdef IS_SERVER
#ifdef IS_CLIENT
#error "Illegal Configuration SERVER-CLIENT"
#else
#ifdef IS_STANDALONE
#error "Illegal Configuration SERVER-STANDALONE"
#endif
#endif
#endif

#ifdef IS_CLIENT
#ifdef IS_STANDALONE
#error "Illegal Configuration CLIENT-STANDALONE"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <math.h>
#include <gmp.h>
#include <BitString.h>
#include <list>
//typedef list<int> TIntList;
#include <set>
#include <vector>


typedef double Tfactor;

struct TSpecialFactorRelation {
  Tfactor factor;
  streampos fpos;

  inline bool operator() (const TSpecialFactorRelation &t1, const TSpecialFactorRelation &t2) const
    { 
      return t1.factor<t2.factor;
    }
};

const char SpecialRelationsFile [] = "special_relations.dat";
const char DynamicRelationsFile [] = "dynamic_relations.dat";
const char StaticRelationsFile []  = "static_relations.dat";
const char RecoveryFile []    = "recovery.dat";

// Fr die Gleichungszeilen zu den Specialfactors
// streams fr das Datenfile zur Verfgung stellen:
filebuf SpecialRelations_buffer;
ostream SpecialRelations_to_file (&SpecialRelations_buffer);
istream SpecialRelations_from_file (&SpecialRelations_buffer);

// Fr die Gleichungszeilen der dynamischen Faktorbasis
// streams fr das Datenfile zur Verfgung stellen:
filebuf DynamicRelations_buffer;
ostream DynamicRelations_to_file (&DynamicRelations_buffer);
istream DynamicRelations_from_file (&DynamicRelations_buffer);

// Fr die Gleichungszeilen der statischen Faktorbasis
// streams fr das Datenfile zur Verfgung stellen:
filebuf StaticRelations_buffer;
ostream StaticRelations_to_file (&StaticRelations_buffer);
istream StaticRelations_from_file (&StaticRelations_buffer);

// Fr Recovery
filebuf Recovery_buffer;
ostream Recovery_to_file (&Recovery_buffer);
istream Recovery_from_file (&Recovery_buffer);

#ifdef IS_CLIENT
char * communication_name;
#endif


// Fr das Speichern der gemachten Faktorisierungen
ofstream Factorization_to_file ("factorizations.txt",ios::in|ios::out|ios::app);
/* Bemerkung:
   Auf diesem Stream werden alle mit diesem Programm gemachten Faktorisierungen
   protokolliert. Die zu faktorisierenden Zahlen sowie alle gefundenen Faktoren
   werden angehngt.
   Das zugehrige File sollte im Bedarfsfall gelscht werden, wenn die Daten
   nicht mehr bentigt werden.
*/

struct TDynamicFactorRelation; // "Vorwrtsdeklaration" fr Headerfile
#include "qsieve.H"

struct TDynamicFactorRelation {
  /* Bemerkung:
     Falls factor nicht allzugro ist (im Augenblick ist dies als Obergrenze der
     int-Bereich), dann kann auch mit den Dynamic-Factors gesiebt werden.
     Nachfolgend die dazu erforderlichen Daten...
  */
  int factor; // factor in int-Darstellung
  int SQRT_kN;
  CGleichungszeile *GL; // Zeiger auf GL (oder Null, falls nicht im Speicher)
  streampos fpos;
  inline bool operator() (const TDynamicFactorRelation &t1, const TDynamicFactorRelation &t2) const
    { 
      return t1.factor<t2.factor;
    }
};


struct TSieb_Delta {
  int delta;
  int factor;
  inline bool operator< (const TSieb_Delta &v) const
    {
#ifdef USE_FIBHEAP
      return (delta < v.delta);
#else
      return (delta > v.delta);
#endif
      // delta > v.delta , if queue is used
      // delta < v.delta , if fibonacci-heap is used
    }
};

#ifdef USE_FIBHEAP
#include "fibheap.H"
FibHeap <TSieb_Delta> Sieb_Delta_Heap, Sieb_Delta_QuadrateHeap;
#else
#include <queue>
priority_queue<TSieb_Delta> Sieb_Delta_Heap, Sieb_Delta_QuadrateHeap;
#endif

const int Specialfactor_Sieb_Treshold  = 1000000000; /* bis hierhin wird auch mit
							Dynamic-Factoren gesiebt */
const int Dynamicfactor_Cache_Treshold =    5000000; /* bis hierhin werden die Dynamic-Factoren
							im Speicher gehalten. Falls der Hauptspeicher
							knapp wird, diesen Wert reduzieren... */

const Tfactor Epsilon_fuer_Factor = 1.0; // fr Feststellung des exakten Fliekommawertes


mpz_t n, /* zu faktorisierende Zahl */
      kN, /* das mit einem gnstigen Vorfaktor multiplizierte n
	     (nachdem n um einfach zu findende Faktoren reduziert wurde) */
      r, /* Quadratwurzel von n (oder bei optimierter Fassung k*n) */
      x,y; /* globale Hilfsvariablen,
	      jede Unterfunktion kann sie nutzen, mu aber davon ausgehen,
	      da sie jederzeit von anderen Funktionen verndert werden! */

int Vorfaktor_von_kN; // der noch zu ermittelnde Vorfaktor fr kN (kN=Vorfaktor_von_kN*n)

const int MAX_Primbasis_size = 50000; /* Maximale Gre der Primzahlbasis */ 
/* Hinweis: Je grer die zu faktorisierende Zahl ist, desto grer
   sollte prinzipiell auch die Primzahlbasis sein, da dann die Relationen
   einfacher gefunden werden knnen.
   Aber: Das lineare Gleichungssystem sprengt bei groer Primzahlbasis
   den Speicher!
   Fr kleinere zu faktorisierende Zahlen wirkt sich dagegen eine grere
   Primzahlbasis verlangsamend aus, da fr ein greres Gleichungssystem
   natrlich auch mehr Relationen gesiebt werden mssen.
   Je nach Geschwindigkeit des Rechners, des verfgbaren Haupt- und
   Hintergrundspeichers (und auch dessen Zugriffszeiten) knnen die
   ttschlich zu verwendenden Werte stark variieren. Sie sind deshalb
   experimentell zu ermitteln.
   Die Maximalgre wird durch MAX_Primbasis_size festgelegt und steht
   damit bereits zur Kompilierzeit fest.
   Bei einer Speicherkapatitt von 32 MB (realem) Hauptspeicher
   hat sich 20000 als brauchbarer Wert erwiesen.
*/

int Primbasis_size = 5000; // tatschlich zu verwendender Wert
/*  Sinnvolle Werte fr Faktorbasen liegen zwischen 1000 und 50000
    (je nach verfgbarem Speicher und der zu faktorisierenden Zahl).
*/

int Primzahlen[MAX_Primbasis_size]; /* fr eine effiziente Probedivision
				       alle Primzahlen der Faktorbasis hier
				       akkumulieren */


int SQRT_kN_der_Primzahlen[MAX_Primbasis_size];
int Delta_der_Primzahlen[MAX_Primbasis_size][2]; /* fr den effizienten Zugriff
						    werden die pro Siebintervall berechneten
						    Werte hier abgelegt */
 
const int FB_max_weitere_Potenzen = 1000; // Anzahl weiterer Primzahlpotenzen, mit denen gesiebt wird
int anz_weitere_Primzahlpotenzen = 0; // tatschliche Anzahl (wird im Programm ermittelt)
int FB_maxQuadrate = 0; /* tatschliche Anzahl der Quadrate,
			   mit denen normal gesiebt wird (wird im Programm ermittelt) */
int Primzahlpotenzen[FB_max_weitere_Potenzen];
int SQRT_kN_der_Primzahlpotenzen[FB_max_weitere_Potenzen];
int Delta_der_Primzahlpotenzen[FB_max_weitere_Potenzen][2];


CGleichungszeile* GLS[MAX_Primbasis_size] = {NULL};
/* Gleichungssystem besteht aus Zeigern auf Gleichungszeilen,
   das mit Null-Pointern initialisiert ist, da anfangs noch
   keine Relationen (=Gleichungszeilen) vorliegen */

CPolynom Polynom; // Objekt zur Verwaltung der Polynomberechnung fr multipolynomiales Sieb



/* Zur Erluterung der Begrifflichkeit von Dynamic-Factors/Special-Factors:
   Ich verwende momentan eine modifizierte Multi-Large-Prime-Variante:
   Dynamic-Factors entsprechen der Large-Prime-Variante. Aber ich siebe mit jedem neu entdeckten
   dynamischen Faktor zustzlich...
   Special-Factors sind zusammengesetzte Restfaktoren (wie sie z.B. in der Double-Large-Prime-Variante
   auftreten). Sie werden in dieser Implementierung allerdings nicht sofort faktorisiert. Stattdessen
   wird lediglich versucht, Special-Factors durch neu aufgenommene Dynamic-Factors zu splitten, um
   daraus weitere Dynamic-Factors zu gewinnen.
   In einer spteren Implementierung knnte man sie auch faktorisiert ablegen und z.B. eine
   Zyklensuche (ggf. auch auf Hyperkantenzyklen) durchfhren.
*/
typedef set<TSpecialFactorRelation,TSpecialFactorRelation> TSpecial_Factor_Relations;
TSpecial_Factor_Relations Special_Factor_Relations;
typedef set<TDynamicFactorRelation,TDynamicFactorRelation> TDynamic_Factor_Relations;
TDynamic_Factor_Relations Dynamic_Factor_Relations;


double Factor_Threshold = 3.0; /* Exponent, der grere Faktoren ber die
				  Faktorbasis hinaus erlaubt.
				  Sinnvolle Werte: 1.0 bis 3.0
				  1.0: keine zustzlichen Faktoren auerhalb der Faktorbasis 
				  bis 2.0: Primfaktoren auerhalb der Faktorbasis bis
				  zum Quadrat des grten Faktors der FB
				  ber 2.0: Faktoren bis zu grter Faktor der Faktorbasis
				  hoch Factor_Threshold.
			       */


const int sieben_ab_Primzahlindex = 5; /* ab diesem Primzahlindex wird gesiebt.
					  Der Wert mu grer als 1 sein, da die ersten beiden Faktoren
					  der Primzahlbasis -1 und mglicherweise die 2 sind. Fr diese
					  beiden Zahlen sowie dem Vorfaktor von kN kann nicht (jedenfalls nicht
					  so einfach) gesiebt werden.
					  Fr ein schnelles, aber unsauberes dirty-sieving kann auch ein
					  grerer Wert gewhlt werden. Aber zu hohe Werte lassen dann auch
					  viele Relationen durchs Sieb fallen... */


const int Siebgroesse = 100000;
/* In Intervallen von MAX_Siebgroesse werden die Relationen gesiebt.
   Hinweis: Je grer MAX_Siebgroesse, desto effizienter geht das Aussieben
   der Relationen. Gleichzeitig erhht sich aber auch der Speicherbedarf.
   Der Einflu der Siebgre ist allerdings geringer als der der gewhlten
   Primzahlbasis. Ein Wert zwischen 1 Mio und 5 Mio drfte daher vollkommen
   ausreichen.
   Andererseits kann mglicherweise im Cachespeicher des Rechners gesiebt
   werden, wenn die Siebgre klein gewhlt wird. Kleine Siebgren erhhen
   zwar den rechnerischen Overhead, aber dafr ist der Zugriff auf das Sieb
   mglicherweise deutlich schneller. Deshalb sollte man auch eine Siebgre
   um 200000 (bei 512K Cache) ausprobieren.
*/

int M = 50000000; // das zu jedem Polynom gehrige Siebintervall ist [-M,M]

int Sieboffset = -M; /* Im allgemeinen wird das fr ein Polynom zu siebende Siebintervall 
			grer sein als die (physikalisch optimale) Siebgre.
			Das Siebintervall ist daher in Teilstcken zu sieben. Hierfr wird der
			Sieboffset bentigt, der den Startpunkt des jeweiligen Teilsiebs angibt. */


typedef short signed int TSiebElement;
TSiebElement* Sieb; // Sieb (noch zu allokierendes Feld von TSiebelementen)
#define log_Multiplikator 100 /* Siebwert wird nach damit
				 nach short int konvertiert */

TSiebElement log_Primzahlen[MAX_Primbasis_size]; /* Hier stehen die konvertierten Logarithmen
						    der Primzahlen aus der Faktorbasis */
TSiebElement log_Primzahl_der_Primzahlpotenzen[FB_max_weitere_Potenzen];
/* fr weitere Primzahlpotenzen stehen hier die konvertierten Logarithmen der Primzahlen */

bool collecting_phase_beendet = false; /* switch, der die "aktive" und "passive" Phase des Siebens
					  bestimmt:
					  Wenn gengend Special-Relations und Dynamic-Relations gesammelt
					  sind, um ausreichend viele Static-Relations zu erhalten, wird
					  das Aufsammeln weiterer Special- und Dynamic-Relations durch
					  das Setzen von collecting_phase_beendet=true unterbunden. */

int gefundene_Relationen = 0;
int Erfolgsaufteilung[5] = {0,0,0,0,0}; /* fr statistische Auswertung, inwieweit die dynamische
					   Factorbase fr das Finden der Relationen
					   verantwortlich ist:
					   [0]->gar nicht beteiligt,
					   [1]-> mit einem Faktor beteiligt
					   [2]-> mit zwei Faktoren beteiligt
					   [3]-> mit drei Faktoren beteiligt
					   [4]-> mit mehr als drei Faktoren beteiligt */

int Special_hit = 0;
int Special_to_dynamic_Factor_hit = 0;

int biggest_Prime_in_Factorbase;


ostream& operator << (ostream& ostr, const mpz_t& value)
{ // Ausgabefunktion fr mpz_t - Zahlen (um die hassenswerten expliziten Konvertierungen zu minimieren...)
  char* str = new char [mpz_sizeinbase(value,10)+2]; mpz_get_str(str, 10, value);
  ostr << str << " (" << strlen(str) << ")";
  delete [] str;              
  return ostr;
}

istream& operator >> (istream& istr, mpz_t& value)
{ // Eingabefunktion fr mpz_t - Zahlen
  string s;
  istr >> s;
  mpz_set_str(value, s.c_str(), 10);
  istr >> s; // Sizeinbase berlesen
  return istr;
}

void tune_parameters(const unsigned int Dezimalstellen)
{
  ifstream in("qsieve.cfg");
  string zeile;
  unsigned int Stellen;
  if (in)
    {
      cout << "Lese Konfigurationsdatei" << endl; 
      while (!in.eof())
	{
	  in >> zeile;
	  
	  Stellen = atoi(zeile.substr(0, zeile.find(":")).c_str());
	  if (Stellen > Dezimalstellen) break;

#ifdef USE_OLD_GCC
	  zeile.remove(0, zeile.find(":")+1);
	  Primbasis_size=atoi(zeile.substr(0, zeile.find(":")).c_str());
	  zeile.remove(0, zeile.find(":")+1);
	  sscanf(zeile.substr(0, zeile.find(":")).c_str(), "%lf", &Factor_Threshold);
	  zeile.remove(0, zeile.find(":")+1);
#else	  
	  zeile.erase(0, zeile.find(":")+1);
	  Primbasis_size=atoi(zeile.substr(0, zeile.find(":")).c_str());
	  zeile.erase(0, zeile.find(":")+1);
	  sscanf(zeile.substr(0, zeile.find(":")).c_str(), "%lf", &Factor_Threshold);
	  zeile.erase(0, zeile.find(":")+1);
#endif

	  M = atoi(zeile.c_str());
	  M = (M>Siebgroesse) ? M : Siebgroesse;
	}
    }
  else cout << "Konfigurationsdatei nicht gefunden!" << endl
	    << "verwende Defaultwerte..." << endl;
  
  cout << "Die verwendeten Parameter fr die Faktorisierung lauten:" << endl;
  cout << "Gre der statischen Faktorbasis: " << Primbasis_size << endl;
  cout << "Factor-Treshold (Exponent): " << Factor_Threshold << endl;
  cout << "Siebintervall pro Polynom: [" << -M << "," << M << "]" << endl;

  if (Primbasis_size>MAX_Primbasis_size)
    {
      cerr << "MAX_Primbasis_size ist zu klein gewhlt (bzw. Primbasis_size zu hoch)!" << endl;
      exit(1);
    }
}


void pollard(const int runden) /* Pollard-rho-Methode */
{
  cout << "Trying Pollard-rho up to " << runden << endl;
  mpz_t a,a2;
  mpz_init(a); mpz_init(a2);
  mpz_set_ui(a,1); mpz_set(a2,a);
  
  for (int i=1; i<=runden; i++)
    {
      mpz_mul(a,a,a); mpz_add_ui(a,a,1); mpz_mod(a,a,n);
      mpz_mul(a2,a2,a2); mpz_add_ui(a2,a2,1); mpz_mod(a2,a2,n);
      mpz_mul(a2,a2,a2); mpz_add_ui(a2,a2,1); mpz_mod(a2,a2,n);
      mpz_sub(x,a2,a);
      mpz_gcd(x,x,n);
      if (mpz_cmp_ui(x,1)!=0)
	{
	  do
	    {
	      if (mpz_probab_prime_p(x,20))
		{
		  cout << x << " ist Faktor. [in Runde " << i <<"]" << endl;
		  Factorization_to_file << x << " * " << flush;
		}
	      else
		{
		  cout << x << " ist zusammengesetzter Faktor. [in Runde " << i <<"]" << endl;
		  Factorization_to_file << x << " [compound]" << " * " << flush;
		}
	      mpz_div(n,n,x);
	      mpz_gcd(x,x,n);
	    } while (mpz_cmp_ui(x,1)!=0);
	  if (mpz_probab_prime_p(n,20)) break; /* Rest wahrscheinlich prim */
	}
    };
  mpz_clear(a); mpz_clear(a2);
}



bool is_prime(register const int p)
{
  if (p==2 || p==3) return true;
  if ((p%2==0) || (p%3==0)) return false;
  register int t = 5;
  while (t*t<=p) 
    {
      if ((p%t)==0) return false;
      t+=2;
      if ((p%t)==0) return false;
      t+=4;
    }
  return true;
}



void pollard_phi(int phase1, int phase2)
{
  // Pollard-Phi Methode modifiziert in zwei Phasen
  
  mpz_t a,b;
  mpz_init(a); mpz_init(b);

  int i;
  short int d;
  
  // Phase 0: kleine Zahlen (mit Primzahlpotenzen)
  cout << "Phi-Phase 0" << endl;
  mpz_set_ui(a,2);
  for (i=5000; i<10000; i++)
    {
      mpz_powm_ui(a,a,i,n);
      mpz_sub_ui(x,a,1); mpz_gcd(x,x,n);
      if (mpz_cmp_ui(x,1)!=0)  /* Teiler gefunden? */
	{
	  cout << "Faktor gefunden in Phase 0: i=" << i << endl;
	  if (mpz_probab_prime_p(x,20))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << x << " * " << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      Factorization_to_file << x << " [compound]" << " * " << flush;
	    }
	  mpz_div(n,n,x);
	  if (mpz_probab_prime_p(n,10)) goto fertig;
	}
    }
	
  // Phase 1: Primzahlen ohne Potenzen (aber knnen in beliebiger Kombination p-1 teilen)
  i=i-(i%6)-1; // i mit Eigenschaft i=5 (mod 6) fr schnellere Primzahlermittlung
  d = 2;
  cout << "Phi-Phase 1" << endl;
  while (phase1>0)
    {
      if (is_prime(i))
	{
	  phase1--;
	  mpz_powm_ui(a,a,i,n);
	  mpz_sub_ui(x,a,1); mpz_gcd(x,x,n);
	  if (mpz_cmp_ui(x,1)!=0)
	    {	  
	      cout << "Faktor gefunden in Phase 1: i=" << i << endl;
	      if (mpz_probab_prime_p(x,20))
		{
		  cout << x << " ist Faktor." << endl;
		  Factorization_to_file << x << " * " << flush;
		}
	      else
		{
		  cout << x << " ist zusammengesetzter Faktor." << endl;
		  Factorization_to_file << x << " [compound]" << " * " << flush;
		}
	      mpz_div(n,n,x);
	      if (mpz_probab_prime_p(n,10)) goto fertig; 
	    }
	}
      i+=d; d=6-d; 
    }


  // Phase 2: isolierte (Prim-)zahl: p-1 mu bis auf diese Zahl vorher vollstndig zerlegtbar gewesen sein.
  cout << "Phi-Phase 2" << endl;
  mpz_set(b,a); // neu Basis fr phi-Methode
  while (!is_prime(i)) { i+=d; d=6-d; }
  mpz_powm_ui(a,a,i,n); mpz_sub_ui(x,a,1);
  
  for (int j=0; j<phase2; j+=1000) 
    { 
      int runden=1000;
      while (runden>0)
	{
	  int i_prev=i; do { i+=d; d=6-d; } while (!is_prime(i));
	  mpz_powm_ui(y,b,i-i_prev,n);
	  mpz_mul(a,a,y); mpz_mod(a,a,n);
	  mpz_sub_ui(y,a,1); mpz_mul(x,x,y); mpz_mod(x,x,n);
	  --runden;
	}
      //cout << "Phase2: " << j << ", " << i << endl;
      mpz_gcd(x,x,n);
      if (mpz_cmp_ui(x,1)!=0)
	{
	  cout << "Faktor gefunden in Phase 2: i=" << i << endl;
	  if (mpz_probab_prime_p(x,20))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << x << " * " << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      Factorization_to_file << x << " [compound]" << " * " << flush;
	    }
	  mpz_div(n,n,x);
	  if (mpz_probab_prime_p(n,10)) goto fertig; 
	}
    }
  
 fertig:
  mpz_clear(a); mpz_clear(b);
}


void fermat_like_methode(unsigned int bis)
{
  /* Fermat-hnliche Methode: wenn t die Zahl n teilt
     und abs(t - n/t) klein ist, so kann hiermit ein Teiler gefunden werden.
     Alternativ ist auch ein beliebiger Vorfaktor fr abs(k*t - n/t)
     realisiert.
     Diese Methode ist zwar nicht besonders effizient, da sie eher einer
     Probedivision entspricht; sie hilft aber, solche Zahlen besonders
     schnell zu faktorisieren, die auf "allzu naive" Weise als
     "besonders schwer" zu faktorisierende Zahlen konstruiert sind.
  */
  cout << "starte Fermat-hnliche Methode" << endl;
  
  mpz_t c,r,h,h2;
  mpz_init(c); mpz_init(r); mpz_init(h); mpz_init(h2);
  
  unsigned int delta;
  
  for (unsigned int k=1; k<=bis; k++)
    {
      cout << "Vorfaktor ist " << k << "   " << endl;
      mpz_mul_ui(c,n,k);
      mpz_mul_ui(c,c,4); // 4c ist notwendig fr "Quadrat-Formel"
      
      mpz_sqrt(r,c);
      for(delta=0; delta<1000000; delta++)
	{
	  if (delta%100000==0)
	    { printf("delta=%i\r",delta); fflush(stdout); }
	  mpz_add_ui(h,r,delta); mpz_mul(h,h,h); mpz_sub(h,h,c);
	  if (mpz_perfect_square_p(h))
	    {
	      { printf("delta=%i\n",delta); fflush(stdout); }
	      cout << "Square detected..." << endl;
	      mpz_sqrt(h2,h); mpz_add_ui(h,r,delta); mpz_sub(h,h,h2);
	      mpz_gcd(h,h,n);
	      if (mpz_cmp_ui(h,1)!=0) 
		{
		  if (mpz_probab_prime_p(h,20))
		    {
		      cout << h << " ist Faktor." << endl;
		      Factorization_to_file << h << " * " << flush;
		    }
		  else
		    {
		      cout << h << " ist zusammengesetzter Faktor." << endl;
		      Factorization_to_file << h << " [compound]" << " * " << flush;
		    }
		  mpz_div(n,n,h);
		  fermat_like_methode(bis); // und noch einmal versuchen...
		  k=bis+1; break;
		}
	    }
	}
    }
  mpz_clear(h); mpz_clear(h2); mpz_clear(r);
  mpz_clear(c);
}



void phimat(mpz_t &n)
{
  mpz_t x,y;
  mpz_init(x); mpz_init(y);
  
  mpz_mul_ui(x,n,4); mpz_sqrt(x,x); // x=trunc(2*sqrt(n));
  mpz_add_ui(y,n,1); mpz_sub(y,y,x); // y=n+1-x
  mpz_set_ui(x,2); mpz_powm(x,x,y,n); // x=2^y mod n;

  cout << "trying phimat..." << x << endl;

  //mpz_add_ui(y,n,1); mpz_div_ui(y,y,2);
  unsigned long int z=0;
  while (z<1000000000)
    {
      int p=mpz_scan1(x,0); z+=p; 
      mpz_tdiv_q_2exp(x,x,p);
      //cout << z << ": " << x << endl;
      if (mpz_cmp_ui(x,1)==0) break;
      mpz_add(x,x,n);
    }

  cout << z << ":" << x << endl;
  
  if (mpz_cmp_ui(x,1)==0)
    {
      cout << "x=1 ... " << endl;
      mpz_mul_ui(x,n,4);mpz_sqrt(x,x); mpz_add_ui(x,x,z); // x=p+q
      mpz_mul(y,x,x);
      mpz_sub(y,y,n); mpz_sub(y,y,n); mpz_sub(y,y,n); mpz_sub(y,y,n);
      mpz_sqrt(y,y); // y=p+q
      mpz_add(y,x,y); mpz_div_ui(y,y,2); // y=p
      mpz_div(x,n,y); // x=q
      mpz_mul(x,x,y);
      cout << x << endl; 
      if (mpz_cmp(x,n)==0) // wirklich echte Teiler ermittelt?
	{
	  mpz_div(x,n,y); // x=q
	  if (mpz_probab_prime_p(x,20))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << x << " * " << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      Factorization_to_file << x << " [compound]" << " * " << flush;
	    }
	  if (mpz_probab_prime_p(y,20))
	    {
	      cout << y << " ist Faktor." << endl;
	      Factorization_to_file << y << " * " << flush;
	    }
	  else
	    {
	      cout << y << " ist zusammengesetzter Faktor." << endl;
	      Factorization_to_file << y << " [compound]" << " * " << flush;
	    }      
	  mpz_set_ui(n,1);
	}
    }
  
  mpz_clear(x); mpz_clear(y);
}



bool k_te_Wurzel (const mpz_t &n, mpz_t &wurzel, const unsigned int k)
{
  /* Berechnen der abgerundeten, ganzzahligen k. Wurzel der bergebenen Zahl.
     Die Wurzel wird in "wurzel" zurckgegeben.
     Der Funktionswert ist true, falls die Zahl die k-te Potenz der zurckgegebenen
     Wurzel ist, andernfalls false. */
  //cout << "k-wurzel rein..." << k << endl;
  mpz_t d,y;
  mpz_init(d); mpz_init(y);

  
  mpz_set_ui(y,1); int exp = 0;
  do { mpz_mul_2exp(y,y,k); exp++; }  while (mpz_cmp(y,n)<=0);
  mpz_set_ui(d,1); mpz_mul_2exp(d,d,exp);
  
  mpz_set_ui(wurzel,0);
  do
    {
      mpz_tdiv_q_2exp(d,d,1);
      mpz_add(wurzel,wurzel,d);
      // mpz_pow_ui(y,wurzel,k); // klappt aus unerfindlichen Grnden nicht, daher:
      mpz_set_ui(y,1); for (int i=k; i>0; i--) mpz_mul(y,y,wurzel); // langsamer Ersatz.

      if (mpz_cmp(y,n)>0) mpz_sub(wurzel,wurzel,d);
    } while (mpz_cmp_ui(d,0)>0);
  
  
  //mpz_pow_ui(y,wurzel,k);
  bool fl=(mpz_cmp(y,n)==0);
  
  mpz_clear(d); mpz_clear(y);
  //cout << "k-wurzel raus..." << endl;
  return fl;
}


bool Potenztest(const mpz_t &n)
{
  // (reine) Potenzen vorhanden?
  mpz_t y;
  mpz_init(y);
  for (unsigned int k=mpz_sizeinbase(n,256); k>1; k--)
    {
      //cout << "k=" << k << "?" << endl;
      if (k_te_Wurzel(n,y,k))
	{	
	  cout << y << "^" << k <<  " ist ";
	  Factorization_to_file << y << "^" << k;
	  if (!mpz_probab_prime_p(y,20))
	    {
	      cout << "zusammengesetzter ";
	      Factorization_to_file << " [compound]";
	    }
	  Factorization_to_file << " * " << flush;
	  cout << "Faktor." << endl;
	  mpz_clear(y);
	  return true; 
	}
    }
  mpz_clear(y);
  return false; // ein Fall fr Viagra...
}


void easy_factor()
{
  /* einfache Faktorisierung versuchen:
     vielleicht kann die Zahl dadurch komplett faktorisiert werden
     oder zumindest um einige Faktoren reduziert werden... */
  
  if (mpz_sizeinbase(n, 10)>50)
    fermat_like_methode(1);
  /* Fermat-hnliche Faktorisierung versuchen, fr den Fall da die Zahl
     z.B. die Form (a-d)*(a+d) mit kleinem d hat... */
  
  cout << "Trialdivision up to 5000" << endl;
  {
    /* Probedivision mit 2 und 3 */
    int exp=0;
    while (mpz_div_ui(y,n,2)==0) { mpz_set(n,y); exp++; }
    if (exp>0)
      {
	cout << "2^" << exp << " ist Faktor." << endl;
	Factorization_to_file << "2^" << exp << " * ";
      }
    exp=0;
    while (mpz_div_ui(y,n,3)==0) { mpz_set(n,y); exp++; }
    if (exp>0)
      {
	cout << "3^" << exp << " ist Faktor." << endl;
	Factorization_to_file << "3^" << exp << " * ";
      }
    
    /* Probedivision bis 5000 */
    unsigned long int p=5;
    while (p<5000)
	{
	  exp=0;
	  while (mpz_div_ui(y,n,p)==0) { mpz_set(n,y); exp++; }
	  if (exp>0)
	    {
	      cout << p << "^" << exp << " ist Faktor." << endl;
	      Factorization_to_file << p << "^" << exp << " * ";
	    }
	  p+=2;
	  
	  exp=0;
	  while (mpz_div_ui(y,n,p)==0) { mpz_set(n,y); exp++; }
	  if (exp>0)
	    {
	      cout << p << "^" << exp << " ist Faktor." << endl;
	      Factorization_to_file << p << "^" << exp << " * ";
	    }
	  p+=4;
	} 
    
    Factorization_to_file.flush();
  }
  
  if (Potenztest(n)) mpz_set_ui(n,1);
  
  if (mpz_cmp_ui(n,1)==0) 
    {
      Factorization_to_file << "1" << endl;
      exit(0); /* 1 als Faktor: komplett faktorisiert */
    }
  
  if (!mpz_probab_prime_p(n,25))
    pollard(150000); /* Faktorisierung nach Pollard-Rho-Methode versuchen */

  if (Potenztest(n)) mpz_set_ui(n,1);
  
  cout << "Phi ist deaktiviert!" << endl;
  //if (!mpz_probab_prime_p(n,25)  && mpz_sizeinbase(n,10)>60 )
  //  pollard_phi(100000,1000000); /* Faktorisierung nach Pollard-Phi-Methode versuchen */
  
  if (!mpz_probab_prime_p(n,25) && mpz_sizeinbase(n, 10)>70)
    fermat_like_methode(5);  /* Fermat-hnliche Faktorisierung versuchen */
  
  cout << "Verbleibende Zahl:" << endl << n << endl;
  
  if ( (mpz_cmp_ui(n,1)==0) || Potenztest(n) )
    {
      Factorization_to_file << "1" << endl;
      exit(0); /* 1 als Faktor: komplett faktorisiert */
    }  
  
  if (mpz_probab_prime_p(n,25))
    {
      Factorization_to_file << n << endl;
      printf("Verbleibende Zahl ist wahrscheinlich prim!\n");
      exit(0);
    }
}



static unsigned int lucas_p, lucas_q, lucas_p_inv;
static int lucas_cache_index = 0;
const int lucas_cache_size = 250;

int lucasv(const int Primzahl, const unsigned int m)
{
  static unsigned int lucas_cache [lucas_cache_size][2];
  
  if (m==0) return 2;
  if (m==1) return lucas_p;
  
  /* Wert schonmal vorher berechnet? */
  for (int i=0; i<lucas_cache_index; i++)
    {
      if (lucas_cache[i][0]==m) return lucas_cache[i][1];
    }
  
  unsigned int wert; /* Rckgabewert fr lucas_v */
  // printf("lucasv(%i)\n",m);
  
  if (m%2==0) 
    {
      unsigned int h1=lucasv(Primzahl,m/2);
      
      mpz_set_ui(y,Primzahl); mpz_set_ui(x,lucas_q);
      mpz_powm_ui(x,x,m/2,y); mpz_mul_ui(x,x,2);
      
      mpz_set_ui(y,h1); mpz_mul(y,y,y);
      
      mpz_sub(y,y,x);
      wert=mpz_mod_ui(y,y,Primzahl);
    }
  else
    {
      unsigned int h1=lucasv(Primzahl,m+1);
      unsigned int h2=lucasv(Primzahl,m-1); 
      mpz_set_ui(x,h2); mpz_mul_ui(x,x,lucas_q);
      mpz_add_ui(x,x,h1);
      mpz_mul_ui(x,x,lucas_p_inv);
      wert=mpz_mod_ui(x,x,Primzahl);
    }
  
  /* Wert in den Cache schreiben... */
  if (lucas_cache_index<lucas_cache_size)
    {
      lucas_cache[lucas_cache_index][0]=m;
      lucas_cache[lucas_cache_index++][1]=wert;
    } else cerr << "Lucas-Cache sollte vergrert werden!" << endl;
  return wert;
}


int lucas(const int Primzahl)
{
  int v;
  if (Primzahl%4!=1)
    { cerr << "Fehler in Lucassequenz Primzahl%4<>1!: " << Primzahl%4 << endl; exit(1); }
  
  /* lucas_p ermitteln */
  lucas_q=mpz_mod_ui(y,kN,Primzahl);
  lucas_p=1;
  do
    {
      lucas_p++;
      mpz_set_ui(x,lucas_p); mpz_mul(x,x,x);
      mpz_set_ui(y,4*lucas_q);
      mpz_sub(y,x,y); mpz_mod_ui(y,y,Primzahl);
      mpz_set_ui(x,Primzahl);
    } while (mpz_legendre(y,x)!=-1);
  
  /* nun das Inverse von lucas_p bezglich Primzahl ermitteln */ 
  mpz_set_ui(x,lucas_p); mpz_set_ui(y,Primzahl); mpz_invert(x,x,y);
  lucas_p_inv=mpz_get_ui(x);
  
  //cout << "Lucas-Cache was " << lucas_cache_index << endl;
  lucas_cache_index=0; /* Cache lschen!! */
  
  /* jetzt die Wurzel von n errechnen */
  v=lucasv(Primzahl,(Primzahl+1)/2);
  /* im folgenden v noch durch 2 teilen (aber modulo Primzahl!!) */
  mpz_set_ui(x,2); mpz_set_ui(y,Primzahl); mpz_invert(x,x,y);
  mpz_set_ui(y,v); mpz_mul(x,x,y);
  v=mpz_mod_ui(y,x,Primzahl);
  return v;
}


int SQRT_kN_mod_Primzahl(const int Primzahl)
{
  /* hier knnte zunchst noch ein Primzahltest erfolgen, doch ich verlasse mich
     aus Effizienzgrnden darauf, da dieser Routine nur Primzahlen bergeben werden! */
  
  if (Primzahl%4==3)
    {
      /* Lsung kann auf einfache Weise effizient berechnet werden... */
      //cout << "effiziente Deltaermittlung: p=3 (mod 4)" << endl;
      mpz_set_ui(x,Primzahl); mpz_powm_ui(y,kN,(Primzahl+1)/4,x);
      return mpz_get_ui(y); 
    }
  else
    if (Primzahl%8==5)
      {
	/* Lsung kann ebenfalls auf einfache Weise effizient berechnet werden... */
	//cout << "effiziente Deltaermittlung: p=5 (mod 8)" << endl;
	mpz_set_ui(x,Primzahl), mpz_powm_ui(y,kN,(Primzahl-1)/4,x);
	if (mpz_get_ui(y)==1)
	  {
	    mpz_powm_ui(y,kN,(Primzahl+3)/8,x);
	    return mpz_get_ui(y);
	  }
	else 
	  {	
	    mpz_set_ui(y,2); mpz_powm_ui(y,y,(Primzahl-1)/4,x);
	    unsigned int w=mpz_get_ui(y);
	    mpz_powm_ui(y,kN,(Primzahl+3)/8,x);
	    mpz_mul_ui(y,y,w);
	    return mpz_mod_ui(y,y,Primzahl);
	  }
      }
    else
      if (Primzahl>100 && (Vorfaktor_von_kN%Primzahl !=0) )
	/* wichtig: keine Lsung durch Lucas-Sequenzen fr Primteiler von kN (und 2)!
	   Auerdem: Fr kleine Zahlen geht Probedivision schneller! */
	{ 
	  /* schnelle Lsung mit Lucas-Sequenzen fr Primzahl=1 (mod 4) */
	  //cout << "Deltaermittlung mit Lucassequenzen: p=1 (mod 4)" << endl;
	  return lucas(Primzahl);
	}
      else
	{
	  /* naive Wurzelsuche (allgemein, aber langsam fr groe Primzahlen!) */
	  //cout << "naive Wurzelsuche" << endl;
	  unsigned int kN_mod_Primzahl = mpz_mod_ui(x,kN,Primzahl); 
	  mpz_set_ui(x,0);
	  do
	    {
	      mpz_mul(y,x,x); // testweise multiplizieren
	      mpz_add_ui(x,x,1); // nchsten Wert vorbereiten
	    } while (mpz_mod_ui(y,y,Primzahl)!=kN_mod_Primzahl);
	  return mpz_get_ui(x)-1;
	}
}


int check_SQRT_kN_mod_Primzahl(const int Primzahl)
{
  // kann zu Debugging-Zwecken aufgerufen werden, um zu testen,
  // inwieweit die "SQRT_kN_mod_Primzahl"-Funktion korrekt arbeitet
  int w1 = SQRT_kN_mod_Primzahl(Primzahl);
  if (w1>Primzahl/2) w1=Primzahl-w1;
  
  mpz_set_ui(x,w1); mpz_mul(x,x,x); mpz_sub(x,x,kN);
  if (mpz_mod_ui(x,x,Primzahl)!=0)
    {
      cerr << "WURZELALARM! " << w1 << " mod " << Primzahl << endl;
      exit(1);
    }
  return w1;
}


void Statische_Faktorbasis_ermitteln()
{ 
  cout << "Ermittle die statische Faktorbasis." << endl;
  Primzahlen[0]=-1; /* nullte "Primzahl" ist -1 (Bercksichtigung des
		       Vorzeichens bei den Relationen! */
  log_Primzahlen[0]=0; // Vorzeichen hat keinen Logarithmuswert
  
  int Primzahl=1;
  
  mpz_t wuqu; mpz_init(wuqu); // Fr Berechnung der Quadratwurzeln von Primzahlpotenzen
  for (int Primzahl_nr=1; Primzahl_nr<Primbasis_size; Primzahl_nr++)
    {
      do // nchsthhere Primzahl der statischen Faktorbasis ermitteln 
	{
	  do Primzahl++; while(!is_prime(Primzahl)); /* nchste Primzahl */
	  mpz_set_ui(x,Primzahl);
	}
      while (mpz_legendre(kN,x)<0); /* Primzahl gltig fr Primzahlbasis? */
      
      //printf("%i. Primzahl der Faktorbasis ist %i\r",Primzahl_nr,Primzahl);
      Primzahlen[Primzahl_nr]=Primzahl;
      
      log_Primzahlen[Primzahl_nr]=static_cast<TSiebElement>(log_Multiplikator*log(Primzahl));
      // fr effizientes Sieben erforderlich (erspart dauernde Neuberechnung)
      
      SQRT_kN_der_Primzahlen[Primzahl_nr]=SQRT_kN_mod_Primzahl(Primzahl);
      // fr sptere Ermittlung der Deltawerte zum Sieben erforderlich
      
      // ggf. Prfung, ob Wurzel korrekt ist:
      mpz_set_ui(x,SQRT_kN_der_Primzahlen[Primzahl_nr]);
      mpz_mul(x,x,x); mpz_sub(x,kN,x); 
      if (mpz_mod_ui(x,x,Primzahl)!=0) { cerr << "Wurzel nicht korrekt!" << endl; exit(1); }
      
      if (Primzahl<=2*M/Primzahl && Primzahl>2 && mpz_mod_ui(x,kN,Primzahl)!=0)
	// Primzahl darf nicht zu gro sein, ungleich 2 und kein Teiler von kN, um mit Potenzen gesiebt zu werden
	{
	  FB_maxQuadrate=Primzahl_nr; /* nach Abschlu der Schleife ist dies der Wert, bis zu dem
					 Quadrate auf normale Weise gesiebt werden. */
	  
	  // Wurzeln von kN modulo Primzahl^pot ermitteln	  
	  // Initialwerte:
	  int P_Potenz = Primzahl*Primzahl;
	  int Wu       = SQRT_kN_der_Primzahlen[Primzahl_nr];
	  
	  for (short int pot=2; ;pot++)
	    {
	      //cout << "berechne Wurzel fr " << Primzahl << "^" << pot << endl;

	      mpz_set_ui(y,2*Wu); mpz_set_ui(x,P_Potenz);
	      if (mpz_invert(y,y,x)==0)
		{ 
		  cerr << "PPot-Wurzel fr " << Primzahl  << ": Inverses existiert nicht!" << endl;
		  break;
		}
	      
	      mpz_mod_ui(x,kN,P_Potenz); mpz_set_ui(wuqu,Wu); mpz_mul_ui(wuqu,wuqu,Wu); mpz_sub(x,x,wuqu);
	      mpz_mul(x,x,y); mpz_add_ui(x,x,Wu); mpz_mod_ui(x,x,P_Potenz);
	      register int h = mpz_get_ui(x); if (h>P_Potenz-h) h=P_Potenz-h; // auf die "kleinste" Wurzel normieren
	      Primzahlpotenzen[anz_weitere_Primzahlpotenzen]=P_Potenz;
	      SQRT_kN_der_Primzahlpotenzen[anz_weitere_Primzahlpotenzen]=h;
	      if ( pot==2 && Primzahl_nr<sieben_ab_Primzahlindex )
		// Exponent 2, weil erstes Auftreten des Faktors
		log_Primzahl_der_Primzahlpotenzen[anz_weitere_Primzahlpotenzen]
		  =static_cast<TSiebElement>(log_Multiplikator*2*log(Primzahl));
	      else
		// Exponent 1, weil nochmaliges Auftreten des Faktors
		// beachte: wenn man schon einmal durch p geteilt hat, ergibt ein weiteres Teilen den Teiler p^2
		log_Primzahl_der_Primzahlpotenzen[anz_weitere_Primzahlpotenzen]
		  =static_cast<TSiebElement>(log_Multiplikator*log(Primzahl));
	      
	      // ggf. Prfung, ob Wurzel korrekt ist:
	      mpz_set_ui(x,SQRT_kN_der_Primzahlpotenzen[anz_weitere_Primzahlpotenzen]);
	      mpz_mul(x,x,x); mpz_sub(x,kN,x); 
	      if (mpz_mod_ui(x,x,P_Potenz)!=0)
		{
		  cerr << "PPot-Wurzel nicht korrekt!" << endl;
		  cerr << "Primzahl=" << Primzahlen[Primzahl_nr] << " Nr. " << Primzahl_nr << endl;
		  cerr << "kN= " << kN << endl;
		  cerr << "SQRT(kN) mod p=" << SQRT_kN_der_Primzahlen[Primzahl_nr] << endl;
		  cerr << "SQRT(kN) mod p^i=" << SQRT_kN_der_Primzahlpotenzen[anz_weitere_Primzahlpotenzen] << endl;
		  cerr << "aber falsch: " << x << endl;
		  exit(1);
		}	  
	      
	      anz_weitere_Primzahlpotenzen++; // dies war eine weitere aufgenommene Primzahlpotenz

	      if (P_Potenz>2*M/Primzahl) break; // nchste Primzahlpotenz wird zu gro...
	      P_Potenz*= Primzahl; Wu=SQRT_kN_der_Primzahlpotenzen[anz_weitere_Primzahlpotenzen-1];
	    }
	  
	}
    }
  mpz_clear(wuqu);

  biggest_Prime_in_Factorbase=Primzahl; 
  cout << "Primzahlen der statischen Faktorbasis ermittelt." << endl;
  
  cout << "Hinweis: ";
  if (Primzahl<Siebgroesse)
    cout << "Alle Primzahlen der statischen Faktorbasis passen in das Sieb!" << endl;
  else
    cout << "Einige Primzahlen der statischen Faktorbasis sind grer als das Sieb!" << endl;
}



void Deltas_berechnen_fuer(const TDynamicFactorRelation &DynRel, const int offset)
{
  TSieb_Delta sieb_delta;
  int d1,d2;
  mpz_t inv_A2, P;
    
  sieb_delta.factor=DynRel.factor;

  mpz_init(inv_A2); mpz_init(P);
  mpz_set_ui(P,DynRel.factor);
  mpz_invert(inv_A2,Polynom.A2,P);

  mpz_set_ui(x,DynRel.SQRT_kN);
  mpz_sub(x,x,Polynom.B); mpz_mul(x,x,inv_A2); 
  d1=mpz_mod_ui(x,x,DynRel.factor);
  sieb_delta.delta=(d1-offset)%DynRel.factor;
  if (sieb_delta.delta<0) sieb_delta.delta+=DynRel.factor;
  sieb_delta.delta+=offset;
  if (sieb_delta.delta<=M) // shortcut: nur dann eintragen, wenn die Relation noch im Siebintervall relevant wird 
    Sieb_Delta_Heap.push(sieb_delta);
  
  mpz_set_ui(x,DynRel.SQRT_kN); mpz_neg(x,x);
  mpz_sub(x,x,Polynom.B); mpz_mul(x,x,inv_A2); 
  d2=mpz_mod_ui(x,x,DynRel.factor);
  sieb_delta.delta=(d2-offset)%DynRel.factor;
  if (sieb_delta.delta<0) sieb_delta.delta+=DynRel.factor;
  sieb_delta.delta+=offset;
  if (d1 != d2)
    {
      if (sieb_delta.delta<=M) // shortcut: nur dann eintragen, wenn die Relation noch im Siebintervall relevant wird 
	Sieb_Delta_Heap.push(sieb_delta);
    }
  mpz_clear(inv_A2); mpz_clear(P);
}


void Special_Factoren_durch_Primfaktor_splitten(const int Primfaktor)
{
  // Wenn ein neuer dynamischer Primfaktor entdeckt wird, so empfiehlt es sich,
  // in der Menge der Specialfaktoren (welches zusammengesetzte Faktoren sind) einen
  // Teilbarkeitstest auf diesen Primfaktor durchzufhren. Auf diese Weise knnen
  // weitere Primfaktoren isoliert werden, die als dynamische Faktoren zur Verfgung stehen.
  
  static int Trashcounter = 0; /* Die (nach Elimination eines Faktors) nutzlos auf dem File
				  befindlichen Relationen */
  
  TSpecial_Factor_Relations::iterator pos = Special_Factor_Relations.begin();
  while (pos!=Special_Factor_Relations.end())
    if (fmod((*pos).factor,Primfaktor)==0.0)
      {
	TSpecialFactorRelation gemerkte_Relation=*pos; // fr korrekte Schleifenbehandlung  erforderlich
	mpz_t x; mpz_init(x); // als lokale Variable erforderlich

	mpz_set_d(x,(*pos).factor);
	if (mpz_div_ui(x,x,Primfaktor)!=0) { cerr << "Unerwartete double-Ungenauigkeit!"; exit(1); }
	
	// zusammenpuzzeln...
	CGleichungszeile *GL = new CGleichungszeile;
	GL->kombinieren(SpecialRelations_from_file,(*pos).fpos);
	SpecialRelations_to_file.seekp((*pos).fpos); SpecialRelations_to_file << "U "; // ungltig markieren

	TDynamicFactorRelation relation;
	relation.factor=Primfaktor;
	TDynamic_Factor_Relations::iterator p = Dynamic_Factor_Relations.find(relation);
	if (p==Dynamic_Factor_Relations.end()) 
	  { 
	    cerr << "Fehler in dynamic factorbase!" << endl; exit(1);
	  }
	mpz_mul_ui(GL->Delta,GL->Delta,(*p).factor); mpz_mod(GL->Delta,GL->Delta,n);
	if ((*p).GL) GL->kombinieren( *((*p).GL) );
	else GL->kombinieren(DynamicRelations_from_file,(*p).fpos);
	
	if (mpz_probab_prime_p(x,20))
	  if (mpz_cmp_ui(x, Specialfactor_Sieb_Treshold)<=0)
	    { 
	      // einen dynamischen Faktor entdeckt!
	      	      
	      // es liegt nun der ermittelte dynamic Factor in x korrekt vor
	      GL->relevanter_Faktor=dynamic_factor;
	      relation.factor=mpz_get_ui(x);
	      
	      // aber ist er neu oder existierte er schon in der dyn. FB?
	      p = Dynamic_Factor_Relations.find(relation);
	      if (p==Dynamic_Factor_Relations.end())
		{ 
		  // er ist neu!
		  //cout << "Special Factor: Dynamischer Faktor erzeugt!" << endl;
		  Special_to_dynamic_Factor_hit++;
		  relation.fpos = GL->speichern(DynamicRelations_to_file,relation.factor);
		  relation.SQRT_kN=SQRT_kN_mod_Primzahl(relation.factor);
		  Deltas_berechnen_fuer(relation, Sieboffset+Siebgroesse);
		  if (relation.factor<=Dynamicfactor_Cache_Treshold) relation.GL=GL; // cachen
		  else { relation.GL=NULL; delete GL; } // nicht cachen
		  Dynamic_Factor_Relations.insert(relation); // Dynamicfactor einfgen
		  Special_Factor_Relations.erase(pos); // alten Specialfactor entfernen
		  Special_Factoren_durch_Primfaktor_splitten(relation.factor);
		}
	      else 
		{ 
		  // er existierte schon in der dyn. FB -> er fhrt zu einem Treffer im GLS!
		  cout << "Special Factor: Special Hit!" << endl;
		  Special_Factor_Relations.erase(pos); // alten Specialfactor entfernen
		  Special_hit++; gefundene_Relationen++; Relation_einfuegen(GL);
		} 
	    }
	  else // x prim und x>SpecialFactorTreshold
	    {
	      // unser x ist zwar nun prim, liegt aber auerhalb des Bereichs fr
	      // die dynamische Faktorbasis -> er ist nutzlos
	      // ---> entfernen
	      cout << "Specialfaktor aus Specialfaktormenge ist nutzlos geworden!" << endl;
	      delete GL;
	      Special_Factor_Relations.erase(pos); // alten Specialfactor entfernen
	    }
	else // x immer noch zusammengesetzt
	  {
	    cout << "Specialfaktor trotz Elimination eines Primfaktors noch zusammengesetzt!" << endl;
	    // es liegt nun ein ermittelter neue special Factor in x korrekt vor
	    GL->relevanter_Faktor=special_factor;
	    
	    TSpecialFactorRelation relation;
	    relation.factor=mpz_get_d(x);
	    
	    // aber ist er neu oder existierte er schon als special Factor?
	    TSpecial_Factor_Relations::iterator p = Special_Factor_Relations.find(relation);
	    if (p==Special_Factor_Relations.end())
	      { 
		// er ist neu!	    
		relation.fpos=GL->speichern(SpecialRelations_to_file,relation.factor);
		Special_Factor_Relations.insert(relation);
		delete GL;
	      }
	    else 
	      {
		// Special-Factor ist bereits bekannt -> er fhrt zu einem Treffer im GLS!
		cout << "Special Hit durch bekannten Special Factor!" << endl;
		Special_hit++; gefundene_Relationen++;
		mpz_mul(GL->Delta,GL->Delta,x); mpz_mod(GL->Delta,GL->Delta,n);
		GL->kombinieren(SpecialRelations_from_file,(*p).fpos);
		Relation_einfuegen(GL);
	      }
	    Special_Factor_Relations.erase(pos); // alten Specialfactor entfernen
	    
	  }
	
	pos=Special_Factor_Relations.upper_bound(gemerkte_Relation); 
	/* Wegen Rekusion etc. wieder korrekt neu aufsetzen, wenn
	   ein Specialfaktor gekegelt wurde -> nchsthherer Special-Factor */
	mpz_clear(x); // lokale Varibale wieder freigeben
	Trashcounter++; // Die auf dem File verbleibenden sinnlosen Relationen zhlen 
      }
    else
      ++pos; // nchsten Special-Factor

  // Wenn sich zuviel Mll angesammelt hat: Mllabfuhr rufen
  if (Trashcounter>50000)
    {
      // Anstatt nur den Mll zu entfernen, machen wir Komplettreset (geht einfacher)
      cout << "Reset der Special-Faktoren" << endl;
      Trashcounter=0;
      //Special_Factor_Relations.clear();
      Special_Factor_Relations.erase(Special_Factor_Relations.begin(),Special_Factor_Relations.end());
      
      SpecialRelations_buffer.close();
      SpecialRelations_buffer.open(SpecialRelationsFile,ios::in|ios::out|ios::trunc);
    }
}



// fr die Speicherung der gefundenen dynamischen Faktoren.
// Bemerkung: Um die dynamischen Faktoren bereits in der Siebphase
//            aufzusammeln, ist das Array global.
//             (Dafr wird die Anzahl der gefundenen dynamischen Faktoren
//              fr die aktuelle Relation dem Konstruktor bergeben.)
const short int MaxHits=50; // bei Bedarf erhhen...
static struct {int Faktor, Exponent;} Hits[MaxHits]; // Dynamische Special-Factoren der Relation aufsammeln
// wichtiger Hinweis: Der Exponent zum Faktor wird nicht beim Sieben, sondern erst bei der Refaktorisierung
// ermittelt. Falls also mit Primzahlpotenzen gesiebt werden sollte, braucht der Exponent nicht eingetragen
// zu werden. Als Faktor sollte in jedem Falle die Primzahl und nicht die Primzahlpotenz eingetragen werden!

CGleichungszeile::CGleichungszeile(const signed int Siebpos, const short int HitCount=0)
  /* Konstruktor: neue Relation einfgen.
     "Deltas" mit dem Wert Delta initialisieren,
     "Relationen" mit den Primfaktoren des reduzierten Quadrates fllen,
     deren Exponenten ungerade sind. */
{
  // cout << "initialisiere Gleichungszeile" << endl;
  Relation_sparse = new CTinyIntVector;
  Relation_dense=NULL;
  mpz_init(Delta);
  Polynom.Werte_holen(Siebpos,Delta,y);
  // in y: reduziertes Quadrat (ist nun zu faktorisieren...) 
  //cout << "y= "; mpz_out_str(stdout,10,y); cout << endl;
  mpz_invert(Delta,Delta,n); /* wir verwenden den Kehrwert von Delta,
				das erspart eine zustzliche Variable;
				gleichzeitig ist damit die zu findende
				quadratische Kongruenz auf Delta^2=1 (mod n)
				normalisiert. */
  
  int z;
  if (mpz_sgn(y)<0) // Vorzeichenbehandlung 
    {
      Relation_sparse->append(0); 
      /* "nullte Primzahl" ist -1 wegen Bercksichtigung
	 des Vorzeichens. Also die 0 fr Position von  -1 einfgen */
      mpz_neg(y,y); /* -1 war Teiler -> positiv machen */
    }
  
  // ersten Faktor der statischen Faktorbasis ebenfalls gesondert behandeln,
  // denn die generierten Polynome haben Probleme, die Primzahl 2 zu sieben...
  z=0;
  while (mpz_div_ui(x,y,Primzahlen[1])==0) { mpz_set(y,x); z++; }
  if (z%2==1) Relation_sparse->append(1); /* ungerade Protenz der i. Primzahl markieren */
  if (z>1) // Delta mit den "gltigen Wurzeln" multiplizieren
  {
    if (z<4) mpz_mul_ui(Delta,Delta,Primzahlen[1]);
    else
      {
	mpz_set_ui(x,Primzahlen[1]); mpz_powm_ui(x,x,z/2,n);
	mpz_mul(Delta,Delta,x);
      }
  }
  
  
  // find factors in dynamic factorbase
  short int HitCount2 = HitCount;
  while (HitCount2--)
    {
      int Exponent=0; // Initialisierung des Exponenten	
      while (mpz_div_ui(x,y,Hits[HitCount2].Faktor)==0)
	{
	  mpz_set(y,x); // getestete Division verwirklichen 
	  Exponent++; // Exponent fr Faktor inkrementieren
	}
      if (Exponent>1)
	{
	  mpz_set_ui(x, Hits[HitCount2].Faktor);
	  mpz_powm_ui(x,x,Exponent/2, n);
	  mpz_mul(Delta,Delta,x); mpz_mod(Delta,Delta,n);
	  Exponent %=2;
	}
      Hits[HitCount2].Exponent = Exponent;
    }
  

  // find factors in stactic factorbase
  // (beginnend mit der 3. Primzahl der FB (s.o) )
  TSiebElement log_w_y = static_cast<TSiebElement>(log_Multiplikator*(log(fabs(mpz_get_d(y)))/2.0));
  int i;
  for (i=2; i<Primbasis_size && log_w_y>=log_Primzahlen[i]; i++)
    {
      
      if ( 
	  ( ((Siebpos-Delta_der_Primzahlen[i][0]) % Primzahlen[i]) !=0)
      	  &&
	  ( ((Siebpos-Delta_der_Primzahlen[i][1]) % Primzahlen[i]) !=0)
	  ) continue; /* keine aufwendige Division durchfhren, wenn diese
			 Primzahl gar nicht ausgesiebt wurde */ 
      
      //if (mpz_div_ui(x,y,Primzahlen[i])!=0) { cerr << "FEHLER!! "; exit(1); }
      
      mpz_div_ui(y,y,Primzahlen[i]); z=1; log_w_y -= log_Primzahlen[i]/2;
      while (mpz_div_ui(x,y,Primzahlen[i])==0) { mpz_set(y,x); z++; log_w_y -= log_Primzahlen[i]/2; }
      if (z%2==1)
	Relation_sparse->append(i); /* ungerade Protenz der i. Primzahl markieren */
      if (z>1) // Delta mit den "gltigen Wurzeln" multiplizieren
	{
	  if (z<4) mpz_mul_ui(Delta,Delta,Primzahlen[i]);
	  else
	    {
	      mpz_set_ui(x,Primzahlen[i]); mpz_powm_ui(x,x,z/2,n);
	      mpz_mul(Delta,Delta,x);
	    }
	  mpz_mod(Delta,Delta,n); // Wert klein halten...
	}
    }
  // clog << i << "/" << Primbasis_size << endl;
  
  // Restfactor in static factorbase...
  bool existiert_grosser_Restfaktor = (mpz_cmp_ui(y,biggest_Prime_in_Factorbase)>0);

  if (existiert_grosser_Restfaktor && collecting_phase_beendet)
    { relevanter_Faktor=special_factor; return; } // vorzeitiges Abbrechen

  if (!existiert_grosser_Restfaktor)
    {
      int p=mpz_get_si(y);
      if (p>1) 
	{
	  while (i<Primbasis_size && p!=Primzahlen[i]) i++;
	  if (i<Primbasis_size)
	    Relation_sparse->append(i); /* Index der Primzahl eintragen, falls
					   Primzahl in Faktorbasis vorhanden ist */
	  else 
	    {
	      cerr << "Fehler: unexpected Factorwert: ";
	      mpz_out_str(stderr,10,y); cerr << " -- " << p << endl;
	    }
	  mpz_set_ui(y,1);
	}
      
      relevanter_Faktor=groesster_Faktor_in_Relation();
      // mit der aufgenommenen Relation lt sich dieser Faktor eliminieren
    }
#ifndef IS_CLIENT  
  // kombiniere Faktoren der dynamic Factorbase
  HitCount2 = HitCount;
  while (HitCount2--) if(Hits[HitCount2].Exponent!=0)
    {
      //cout << "kombiniere dynamischen Factor " << Hits[HitCount].Faktor << endl;
      TDynamicFactorRelation FaRelSearch;
      FaRelSearch.factor=Hits[HitCount2].Faktor;
      TDynamic_Factor_Relations::iterator p = Dynamic_Factor_Relations.find(FaRelSearch);
      if (p==Dynamic_Factor_Relations.end()) cerr << "Fehler in dynamic factorbase!" << endl;
      mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
      if ((*p).GL) kombinieren( *((*p).GL) );
      else kombinieren(DynamicRelations_from_file,(*p).fpos);
      //cout << "fertig kombiniert von " << *pos << endl;
    }
#endif
  if (existiert_grosser_Restfaktor)
    {
      if (mpz_probab_prime_p(y,20))
	if (mpz_cmp_ui(y, Specialfactor_Sieb_Treshold)<=0)
	  {
	    // es liegt ein neuer dynamic Factor vor
	    //cout << "neuer dynamic-factor..." << endl;
	    relevanter_Faktor=dynamic_factor;
	    TDynamicFactorRelation relation;
	    relation.factor=mpz_get_ui(y);

#ifndef IS_CLIENT
	    relation.fpos = speichern(DynamicRelations_to_file,relation.factor);
	    if (relation.factor<Dynamicfactor_Cache_Treshold)
	      {
		relation.GL=this; relevanter_Faktor=cached_dynamic_factor;
	      }
	    else
	      {
		relation.GL=NULL;
	      }
#else
#ifdef USE_NETWORK
	    unix_io_client communication_stream(communication_name, server_port);
#else
	    ofstream communication_stream(communication_name, ios::out|ios::app);
#endif
	    // an den server
	    communication_stream << "Relation! ";
	    relation.fpos=speichern(communication_stream,relation.factor, HitCount);
#endif
	    relation.SQRT_kN=SQRT_kN_mod_Primzahl(relation.factor);
	    Deltas_berechnen_fuer(relation, Sieboffset+Siebgroesse);
	    Dynamic_Factor_Relations.insert(relation);
#ifndef IS_CLIENT
	    Special_Factoren_durch_Primfaktor_splitten(relation.factor);
#endif
	    return;
	  }
	else 
	  {
	    // es liegt ein Special-Factor vor, den wir aber ignorieren,
	    // weil er prim ist
	    relevanter_Faktor=special_factor;
	    return;
	  }
      else // es liegt ein Specialfactor vor, der nicht prim ist
	{
	  relevanter_Faktor=special_factor;
	  TSpecialFactorRelation relation;
	  relation.factor=mpz_get_d(y);
	  
	  volatile Tfactor f = relation.factor+Epsilon_fuer_Factor;
	  if (f==relation.factor) // reicht die Fliekommagenauigkeit?
	    {
	      cerr << "Hinweis: Special Factor " << y << " bersteigt Fliekommagenauigkeit!" << endl;
	      return; /* Treffer nicht bernehmen, da er den double-Bereich
			 in der Mantissengenauigkeit berschreitet! */
	    }
#ifndef IS_CLIENT	  
	  TSpecial_Factor_Relations::iterator pos = Special_Factor_Relations.find(relation);
	  if (pos==Special_Factor_Relations.end())
	    {
	      relation.fpos=speichern(SpecialRelations_to_file,relation.factor);
	      Special_Factor_Relations.insert(relation);
	    }
	  else
	    { 
	      // Special-Factor ist bereits bekannt
	      Special_hit++;
	      gefundene_Relationen++;
	      mpz_mul(Delta,Delta,y); mpz_mod(Delta,Delta,n);
	      kombinieren(SpecialRelations_from_file,(*pos).fpos);
	    }
#else
#ifdef USE_NETWORK
	  unix_io_client communication_stream(communication_name, server_port);
#else
	  ofstream communication_stream(communication_name, ios::out|ios::app);
#endif
	  // an den server
	  communication_stream << "Relation! ";
	  relation.fpos=speichern(communication_stream,relation.factor, HitCount);
#endif
	  return;
	}
    }

#ifdef IS_CLIENT
#ifdef USE_NETWORK
      unix_io_client communication_stream(communication_name, server_port);
#else
      ofstream communication_stream(communication_name, ios::out|ios::app);
#endif
      communication_stream << "Relation! ";
      speichern(communication_stream,1, HitCount);
#endif

  gefundene_Relationen++;
  if (HitCount<4) Erfolgsaufteilung[HitCount]++; else Erfolgsaufteilung[4]++;
  // bei Erfolgsaufteilung[0]: Relation vollstndig durch statische FB entdeckt
  // bei Erfolgsaufteilung[1]: Relation durch Eliminierung eines Faktors aus dyn. FB entdeckt
  // bei Erfolgsaufteilung[2]: Relation durch Eliminierung zweier Faktoren aus dyn. FB entdeckt
  // bei Erfolgsaufteilung[3]: Relation durch Eliminierung dreier Faktoren aus dyn. FB entdeckt
  // bei Erfolgsaufteilung[4]: Relation durch Eliminierung >3 Faktoren aus dyn. FB entdeckt
  
  //cout << "GL: " << *this << endl;
  //cout << "relevanter Faktor: " << relevanter_Faktor << endl;
  //cout << "Initialisierung der Gleichungszeile abgeschlossen!" << endl;
}

void CGleichungszeile::Ausgabe() const
{
  if (Relation_dense) 
    cout << Relation_dense->count(1) << " Primes (dense) ";
  else 
    for (int i=0; i<Relation_sparse->size(); i++) cout << Relation_sparse->v[i] << " ";
  
  cout << "<=> ";
  mpz_out_str(stdout,10,Delta);
  cout << endl;
}


ostream& operator<< (ostream& ostr, const CGleichungszeile& GL)
{
  if (GL.Relation_dense) ostr << "dense"; else ostr << "sparse";
  ostr << "1=";
  
  char *str;
  str = new char [mpz_sizeinbase(GL.Delta,10)+2]; mpz_get_str(str, 10, GL.Delta);
  ostr << str << "^2";
  delete [] str; 
  
  if (GL.Relation_dense)
    { 
      for (int i=GL.Relation_dense->first(); i>=0; i=GL.Relation_dense->next(i))
	ostr << "*p" << i;
    }
  else
    {
      for (int pos=0; pos<GL.Relation_sparse->size(); ++pos)
	ostr << "*p" << GL.Relation_sparse->v[pos];
    }
  
  return ostr;
}


void CGleichungszeile::convert_Relation_to_dense()
{
  //cout << "Konvertiere von Sparsedarstellung nach Densedarstellung." << endl;
  if (Relation_dense)
    { cerr << "Relation ist bereits dense!" << endl; return; }
  Relation_dense = new BitString;
  for (int pos=0; pos<Relation_sparse->size(); ++pos)
    Relation_dense->set(Relation_sparse->v[pos]);
  //Relation_sparse->erase(Relation_sparse->begin(),Relation_sparse->end());
  delete Relation_sparse; Relation_sparse=NULL; // Sparsedarstellung lschen
}

void CGleichungszeile::convert_Relation_to_sparse()
{
  //cout << "Konvertiere von Densedarstellung nach Sparsedarstellung." << endl;
  if (Relation_sparse)
    { cerr << "Relation ist bereits sparse!" << endl; return; }
  Relation_sparse = new CTinyIntVector(Relation_dense->count(1));
  for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
    Relation_sparse->fast_append(i);
  //Relation_dense->clear();
  delete Relation_dense; Relation_dense=NULL; // Densedarstellung lschen
}


void CGleichungszeile::check_Relation(const Tfactor Restfaktor) const
{
  mpz_t h;
  mpz_init(h);
  mpz_mul(h,Delta,Delta); mpz_mod(h,h,n);
  
  if (Relation_dense)
    { 
      for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
	{
	  if (i>0) mpz_mul_ui(h,h,Primzahlen[i]); else mpz_neg(h,h);
	  mpz_mod(h,h,n);
	}
    }
  else
    {
      for (int pos=0; pos<Relation_sparse->size(); ++pos)
	{
	  if (Relation_sparse->v[pos]>0) mpz_mul_ui(h,h,Primzahlen[Relation_sparse->v[pos]]);
	  else mpz_neg(h,h);
	  mpz_mod(h,h,n);
	}
    }
  

  mpz_t h2;
  mpz_init(h2);
  mpz_set_d(h2,Restfaktor);
  mpz_mul(h,h,h2); mpz_mod(h,h,n);
  mpz_clear(h2);
  
  if (mpz_cmp_ui(h,1)!=0)
    {
      cerr << "Relation mit Faktor " << setprecision(20) << Restfaktor
	   << " ist nicht korrekt: 1<>" << h << endl;
      exit(1);
    }
  else   
    {
      cerr << "Relation mit Faktor " << setprecision(20) << Restfaktor
	   << " ist korrekt: 1=" << h << endl;
    }
  
  mpz_clear(h);
}


void CGleichungszeile::kombinieren(const CGleichungszeile& GL2)
  // bergabe von GL2 ber Referenz, um effizienter zu sein...
  /* symmetrische Differenzen ermitteln */
{
  //cout << "Kombiniere Gleichungszeilen..." << endl;
  //cout << "Gl.1: "; Kongruenzcheck();
  //cout << "Gl.2: "; GL2.Kongruenzcheck();
  
  /*
    Symmetrische Differenz der Factors in Relation bilden.
    Gemeinsamen Anteil in einfacher(!!) Potenz in Relationswurzel aufnehmen
  */
  mpz_set_ui(x,1); // zur Beschleunigung Faktoren in x aufsammeln, bis sie grer als n werden

  if ( GL2.Relation_dense ) /* GL2.Relation is dense */
    {
      if (!Relation_dense) // Relation ist nicht dense
	convert_Relation_to_dense(); // Relation in dichte Darstellung bringen
      
      for (int i=GL2.Relation_dense->first(); i>=0; i=GL2.Relation_dense->next(i))
	if (Relation_dense->test(i))
	  if (i>0)
	    {
	      mpz_mul_ui(x,x,Primzahlen[i]);
	      if (mpz_cmp(x,n)>=0)
		{
		  mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		  mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		}
	    } //else mpz_neg(Delta,Delta);
      
      (*Relation_dense)^=*GL2.Relation_dense; // symmetrische Differenz bilden
    }
  else /* GL2.Relation is sparse */
    if (Relation_dense) /* and Relation is dense */
      { /* symmetrische Differenz in dense-Darstellung bilden,
	   wobei der andere Vektor sparse ist */
	for (int pos=0; pos<GL2.Relation_sparse->size(); ++pos)
	  {
	    if (Relation_dense->test(GL2.Relation_sparse->v[pos]))
	      if (GL2.Relation_sparse->v[pos]>0)
		{
		  mpz_mul_ui(x,x,Primzahlen[GL2.Relation_sparse->v[pos]]);
		  if (mpz_cmp(x,n)>=0)
		    {
		      mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		      mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		    }
		} //else mpz_neg(Delta,Delta);
	    Relation_dense->invert(GL2.Relation_sparse->v[pos]);
	  }
      }
    else
      { /* symmetrische Differenz in sparse-Darstellung bilden */
	int p1 = 0;
	int p2 = 0;
	static CTinyIntVector neueMenge(Primbasis_size);
	neueMenge.reset();
	while (p2<GL2.Relation_sparse->size() && p1<Relation_sparse->size())
	  {
	    while (GL2.Relation_sparse->v[p2]<Relation_sparse->v[p1])
	      {
		neueMenge.fast_append(GL2.Relation_sparse->v[p2]); ++p2;
		if (p2==GL2.Relation_sparse->size()) goto fertig;
	      }
	    while (Relation_sparse->v[p1]<GL2.Relation_sparse->v[p2])
	      {
		neueMenge.fast_append(Relation_sparse->v[p1]); ++p1;
		if (p1==Relation_sparse->size()) goto fertig;
	      }
	    while (GL2.Relation_sparse->v[p2]==Relation_sparse->v[p1])
	      {
		if (Relation_sparse->v[p1]>0)
		  {
		    mpz_mul_ui(x,x,Primzahlen[Relation_sparse->v[p1]]);
		    if (mpz_cmp(x,n)>=0)
		      {
			mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
			mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		      }
		  } //else mpz_neg(Delta,Delta);
		++p1; ++p2;
		if (p2==GL2.Relation_sparse->size()) goto fertig;
		if (p1==Relation_sparse->size()) goto fertig;      
	      }
	  }
      fertig:
	while (p2!=GL2.Relation_sparse->size()) { neueMenge.fast_append(GL2.Relation_sparse->v[p2]); ++p2; }
	while (p1!=Relation_sparse->size()) { neueMenge.fast_append(Relation_sparse->v[p1]); ++p1; }
	Relation_sparse->copy_from(neueMenge.v,neueMenge.size()); // neue Menge bernehmen
	
	if (Relation_sparse->size()>Primbasis_size/50)
	  /* Schwelle fr dense berschritten? */
	  {
	    convert_Relation_to_dense(); /* dann von sparse nach dense
					    konvertieren */
	  }
	else Relation_sparse->optisize(); // optimize Size of vector
      }
  
  // x beinhaltet (mglicherweise) noch einen Restwert, der zu Delta multipliziert werden mu:
  mpz_mul(Delta,Delta,x); mpz_mod(Delta,Delta,n);
  
  // nicht vergessen: Das neue Delta beinhaltet natrlich auch das
  // alte Delta der anderen Gleichungszeile; diese also dazumultiplizieren!
  // Relationswurzeln beider Gleichungszeilen, also dazumultiplizieren:
  mpz_mul(Delta,Delta,GL2.Delta); mpz_mod(Delta,Delta,n);
  
  
  relevanter_Faktor=groesster_Faktor_in_Relation();
  //cout << "Kombinierung beendet." << endl;
}

Tfactor CGleichungszeile::kombinieren(istream &in, const streampos pos)
{
  in.seekg(pos);
  return kombinieren(in);
}

Tfactor CGleichungszeile::kombinieren(istream &in)
{
  //cout << "kombiniere Gleichungszeile vom File" << endl;
  Tfactor ret;
  string s;
  int index;
  
  in >> s; // Startzeichen
  if (s != "G")
    {
      cerr << "Error: Falsche Position beim Einlesen der Gleichungszeile" << endl;
      exit(1);
    }
  in >> ret; // Faktor einlesen (dynamic-factor oder special-factor der Relation oder 1)
  in >> s; // delta
  mpz_set_str(x, s.c_str(), 10); // Thorsten: "x ist frei"
  mpz_mul(Delta, Delta, x);
  mpz_mod(Delta, Delta, n);
  
  /*
    Symmetrische Differenz der Factors in Relation bilden.
    Gemeinsamen Anteil in einfacher(!!) Potenz in Relationswurzel aufnehmen
  */
  mpz_set_ui(x,1); // zur Beschleunigung Faktoren in x aufsammeln, bis sie grer als n werden
  int distance; // Hinweis: statt der Werte werden deren Abstnde eingelesen...

  if (Relation_dense) /* Relation is dense */
    { /* symmetrische Differenz in dense-Darstellung bilden */
      in >> distance; index=distance;
      while (distance >= 0)
	{
	  if (Relation_dense->test(index))
	    if (index>0)
	      {
		mpz_mul_ui(x,x,Primzahlen[index]);
		if (mpz_cmp(x,n)>=0)
		  {
		    mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		    mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		  }
	      } //else mpz_neg(Delta,Delta);
	  Relation_dense->invert(index);
	  in >> distance; index+=distance;
	}
    }
  else
    { /* symmetrische Differenz in sparse-Darstellung bilden */
      int p1 = 0;
      static CTinyIntVector neueMenge(Primbasis_size);
      neueMenge.reset(); // "leeren" ohne Speicherfreigabe
      
      in >> distance; index=distance; 
      while (distance>=0 && p1<Relation_sparse->size())
	{
	  while (index<Relation_sparse->v[p1])
	    {
	      neueMenge.append(index); in >> distance; index+=distance;
	      if (distance<0) goto fertig;
	    }
	  while (Relation_sparse->v[p1]<index)
	    {
	      neueMenge.append(Relation_sparse->v[p1]); ++p1;
	      if (p1==Relation_sparse->size()) goto fertig;
	    }
	  while (index==Relation_sparse->v[p1])
	    {
	      if (index>0)
		{
		  mpz_mul_ui(x,x,Primzahlen[index]);
		  if (mpz_cmp(x,n)>=0)
		    {
		      mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		      mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		    }
		} //else mpz_neg(Delta,Delta);
	      in >> distance; index+=distance; ++p1;
	      if (distance<0) goto fertig;
	      if (p1==Relation_sparse->size()) goto fertig;      
	    }
	}
    fertig:
      while (distance>=0) { neueMenge.append(index); in >> distance; index+=distance; }
      while (p1!=Relation_sparse->size()) { neueMenge.append(Relation_sparse->v[p1]); ++p1; }
      Relation_sparse->copy_from(neueMenge.v,neueMenge.size()); // neue Menge bernehmen
      
      if (Relation_sparse->size()>Primbasis_size/50)
	/* Schwelle fr dense berschritten? */
	{
	  convert_Relation_to_dense(); /* dann von sparse nach dense
					  konvertieren */
	}
      else Relation_sparse->optisize(); // optimize Size of vector
    }
  
  // x beinhaltet (mglicherweise) noch einen Restwert, der zu Delta multipliziert werden mu:
  mpz_mul(Delta,Delta,x); mpz_mod(Delta,Delta,n);
 
  if (distance == -2) // Korrekturfaktoren
    {
      /* Korrekturfaktoren sind dynamische Faktoren, die zum Zeitpunkt der Generierung
	 der gelesenen Relation nicht aufgelst wurden. Sie mssen daher jetzt verknpft werden. */
      TDynamicFactorRelation FaRelSearch;
      // Korrekturfaktoren
      in >> FaRelSearch.factor;
      while (FaRelSearch.factor>0)
	{
	  //cout << "Korrekturfaktor: " << FaRelSearch.factor << endl;
	  //cout << "kombiniere dynamischen Factor " << Hits[HitCount].Faktor << endl;
	  TDynamic_Factor_Relations::iterator p = Dynamic_Factor_Relations.find(FaRelSearch);
	  if (p==Dynamic_Factor_Relations.end()) cerr << "Fehler in dynamic factorbase!" << endl;
	  mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
	  if ((*p).GL) kombinieren( *((*p).GL) ); // dynamische Relation im Hauptspeicher gecached
	  else kombinieren(DynamicRelations_from_file,(*p).fpos); // auf Hintergrundspeicher
	  //cout << "fertig kombiniert von " << *pos << endl;
	  in >> FaRelSearch.factor;
	}
    }
  
  relevanter_Faktor=groesster_Faktor_in_Relation();
  //cout << "Kombinierung beendet." << endl;
  return ret;
}

streampos CGleichungszeile::speichern(ostream &out, const Tfactor factor=1, const short int HitCount = 0) const
{
  //cout << "Speichern einer Gleichungszeile" << endl;
#ifndef IS_CLIENT
  //check_Relation(factor); // Test, ob die Kongruenz auch aufgeht (sollte sie nmlich!)
  /* Der Check ist nur erforderlich, wenn sichergestellt werden soll, da
     der Server nicht durch falsche Relationen defektiert werden kann.
     (Oder zu Debuggingzwecken ;)
     WICHTIG: Der Check klappt (wegen der Korrekturfaktoren) nicht im Clientmodus!
  */
  
  streampos pos;
  out.seekp(0,ios::end); pos = out.tellp(); // fileposition ans Ende setzen und auslesen
#endif
  char *str;

  str = new char [mpz_sizeinbase(Delta,10)+2]; mpz_get_str(str, 10, Delta);
  out << "G " << setprecision(20) << factor << " " << str << " ";
  delete [] str;
  
  /* Bemerkung:
     Statt der Werte werden die Abstnde gespeichert, das spart Platz */
  if (Relation_dense)
    { 
      int prev = 0;
      for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
	{
	  out << i-prev << " ";
	  prev=i;
	} 
    }
  else
    {
      int prev = 0;
      for (int i=0; i<Relation_sparse->size(); ++i)
	{
	  out << Relation_sparse->v[i] - prev << " ";
	  prev=Relation_sparse->v[i];
	}
    }
#ifndef IS_CLIENT
  out << "-1" << endl;
  return pos;
#else
  int Anzahl_echte_Faktoren = 0;
  for (int i=0; i<HitCount; i++) if (Hits[i].Exponent>0) Anzahl_echte_Faktoren++;
  if (Anzahl_echte_Faktoren>0)
    {
      out << "-2" << " ";
      //cout << "Korrekturfaktoren fr Relation mit Faktor " << setprecision(20) << factor << endl;
      for (int i=0; i<HitCount; i++)
	{
	  //cout << Hits[i].Faktor << "^" << Hits[i].Exponent << endl; 
	  if (Hits[i].Exponent>0) out << Hits[i].Faktor << " ";
	}
    }
  out << "-1" << endl;
  return 0;
#endif
}

void CGleichungszeile::Kongruenzcheck() const
{
  mpz_set_ui(y,1); 
  if (Relation_sparse)
    {
      for (int i=0; i<Relation_sparse->size(); ++i)
	{
	  if (Relation_sparse->v[i]>0) 
	    mpz_mul_ui(y,y,Primzahlen[Relation_sparse->v[i]]);
	  else mpz_neg(y,y);
	  mpz_mod(y,y,n);
	}
    }
  else
    {
      for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
	{
	  if (i>0) mpz_mul_ui(y,y,Primzahlen[i]); else mpz_neg(y,y);
	  mpz_mod(y,y,n);
	}
    }
  
  mpz_invert(x,Delta,n); mpz_powm_ui(x,x,2,n);
  
  cout << "Kongruenzcheck: ";
  mpz_out_str(stdout,10,x); printf(" = ");
  mpz_out_str(stdout,10,y); printf(" (mod N)\n");
  if (mpz_cmp(x,y)!=0) 
    {
      cout << "Fehler! Werte sind NICHT kongruent!" << endl;
      exit(1);
    }
  
}


void CGleichungszeile::Quadratische_Kongruenz_ausrechnen() const
{
  /* Die ermittelte Gleichungszeile erfllt die Kongruenz (A=1) mod n.
     Wegen A-1=(A+1)*(A-1)=0 mod n knnen wir daher nun durch ggT(A+1,n)
     einen Teiler von n ermitteln und haben gute Chancen, da es sich dabei
     auch um einen echten (also nichttrivialen) Teiler handelt... */

  cout << endl << "Geeignete Relation ermittelt..." << endl;

  if (!Relation_ist_leer()) cout << "Fehler! Relation ist noch nicht leer!" << endl;

  cout << Delta << " = 1 (mod N)" << endl;
   
  mpz_powm_ui(x,Delta,2,n);
  printf("d.h.: ");
  cout << x << " = 1 (mod N)" << endl;
  
  if (mpz_cmp_ui(x,1)!=0) 
    {
      cout << "Fehler! Quadrate sind NICHT kongruent!" << endl;
      exit(1);
    }
  
  mpz_add_ui(x,Delta,1);
  mpz_gcd(x,x,n); mpz_div(y,n,x); /* beachte: n und nicht kN
				     ist die eigentlich zu
				     faktorisierende Zahl! */
  
  if ((mpz_cmp_ui(x,1)==0) || (mpz_cmp_ui(y,1)==0))
    {
      cout << "Faktorisierung war leider trivial..." << endl;
    }
  else
    {
      cout << endl << "Faktorisierung gefunden!" << endl;
      cout << x << " * ";
      cout << y << endl;
      cout << "= ";
      cout << n << endl;
      
      if (mpz_probab_prime_p(x,25))
	{
	  mpz_div(n,n,x);
	  cout << x << " ist Faktor." << endl;
	  Factorization_to_file << x << " * " << flush;
	}
      else
      	if (Potenztest(x)) mpz_div(n,n,x); // ggf. Potenzen streichen...

      if (mpz_probab_prime_p(y,25))
	{
	  mpz_div(n,n,y);
	  cout << y << " ist Faktor." << endl;
	  Factorization_to_file << y << " * " << flush;
	}
      else
	if (Potenztest(y)) mpz_div(n,n,y); // ggf. Potenzen streichen...
      

      if (mpz_cmp_ui(n,1)==0) // Faktorisierung vollstndig?
	{ 
	  cout << "Faktorisierung erfolgreich beendet." << endl;
	  Factorization_to_file << "1" << endl;
	  
	  // Streams schlieen
          SpecialRelations_buffer.close();
	  DynamicRelations_buffer.close();
	  StaticRelations_buffer.close();
	  Recovery_buffer.close();

	  // Files lschen
	  remove(SpecialRelationsFile);
	  remove(DynamicRelationsFile);
	  remove(StaticRelationsFile);	
	  remove(RecoveryFile);
	  
	  exit(0); // Programm erfolgreich beenden...
	}
    }
  cout << "Setze Faktorisierung fort..." << endl;
}
  
unsigned int CGleichungszeile::SizeOfRelation_verknuepft_mit(const CGleichungszeile& GL2) const
{
  int Zaehler = 0;
  
  if (Relation_dense)
    if (GL2.Relation_dense)
      {
	for (int i=GL2.Relation_dense->first(); i>=0; i=GL2.Relation_dense->next(i))
	  if (Relation_dense->test(i)) Zaehler++;
      }
    else 
      {
        for (int i=0; i<GL2.Relation_sparse->size(); ++i)
	  if (Relation_dense->test(GL2.Relation_sparse->v[i])) Zaehler++;
      }
  else // Relation is sparse
    if (GL2.Relation_dense)
      {
	for (int i=0; i<Relation_sparse->size(); ++i)
	  if (GL2.Relation_dense->test(Relation_sparse->v[i])) Zaehler++;
      }
    else
      {
	int j=0;
	for (int i=0; i<GL2.Relation_sparse->size(); ++i)
	  {
	    while (j<Relation_sparse->size()
		   && Relation_sparse->v[j]<GL2.Relation_sparse->v[i]) ++j;
	    if (j>=Relation_sparse->size()) break;
	    if (Relation_sparse->v[j]==GL2.Relation_sparse->v[i]) {j++; Zaehler++;}
	  }
      }
  
  return SizeOfRelation()+GL2.SizeOfRelation()-2*Zaehler;
}


void CGleichungszeile::Relation_optimieren()
{      
  int bitpos = 1;
  while (true)
    {
      signed int vorletzter_Faktor;
      if (Relation_sparse) 
	if (Relation_sparse->size()>bitpos)
	  vorletzter_Faktor=Relation_sparse->v[Relation_sparse->size()-1-bitpos];
	else break;
      else
	{
	  vorletzter_Faktor=Relation_dense->last(1);
	  for (int j=0; j<bitpos && vorletzter_Faktor>=0; j++)
	    vorletzter_Faktor=Relation_dense->prev(vorletzter_Faktor,1);
	  if (vorletzter_Faktor<0) break;
	}
      if (GLS[vorletzter_Faktor]==NULL) { bitpos++; continue; }
      // cout << "Optimiere " << groesster_Faktor_in_Relation()
      //	   << " mit " << vorletzter_Faktor << endl; 
      kombinieren(*GLS[vorletzter_Faktor]);
    }
  
  if (Relation_dense && Relation_dense->count(1)<Primbasis_size/50)
    {
      // Relation nach sparse konvertieren
      convert_Relation_to_sparse();
    }
}


void Gleichungssystem_optimieren()
{
  /* Mglichst viele redundante Faktoren aus den Relationen entfernen.
     Dadurch dafr sorgen, da die Relationen mglichst dnn mit Faktoren
     besetzt sind.
     -> Nachteil: bei dicht besetztem Gleichungssystem hohe Laufzeit! */
  cout << "Optimiere Gleichungssystem" << endl;
  for (int i=1; i<Primbasis_size; i++) if (GLS[i]) 
    {
      cout << " GL " << i << ", " << flush;
      GLS[i]->Relation_optimieren();
    }
  cout << endl;
}



void Statusmeldung()
{
  // aktuellen Stand ausgeben  
  cout << gefundene_Relationen
       << "{" << Erfolgsaufteilung[0]
       << "," << Erfolgsaufteilung[1]
       << "," << Erfolgsaufteilung[2]
       << "," << Erfolgsaufteilung[3]
       << "," << Erfolgsaufteilung[4]
       << "}/" << Primbasis_size << ","
       << Special_Factor_Relations.size()
       << "(" << Special_to_dynamic_Factor_hit << "," << Special_hit << ")"
       << "/" << Dynamic_Factor_Relations.size()
       << " [" << Sieboffset << "] -> ";
  cout.setf(ios::left, ios::adjustfield); 	
  cout << "ca. " << setw(6) << setprecision(5)
       << 100.0*gefundene_Relationen/Primbasis_size
       << "%  \n" << flush; 
}


void Relation_einfuegen(CGleichungszeile *GL)
{ /* hier wird eine neu gefundene Relation in das bisherige Gleichungssystem
     eingefgt. Falls sie redundant ist bzw. zu einer Faktorisierung fhrt,
     so wird dies automatisch erkannt. */
#ifdef IS_CLIENT
  delete GL;
  return;
#else  
  if (GL->relevanter_Faktor==dynamic_factor || GL->relevanter_Faktor==special_factor)
    { delete GL; return; } /* diese Relationen sind auf Hintergrundspeicher ausgelagert
			      und brauchen deshalb nicht im Hauptspeicher zu verbleiben */
  
  if (GL->relevanter_Faktor==cached_dynamic_factor)
    return; /* Diese Faktoren sind zwar auch auf dem Hintergrundspeicher, verbleiben aber
	       zwecks schnelleren Zugriffs auch im Hauptspeicher */
  
  while (!GL->Relation_ist_leer())
    {
      //cerr << "kombiniere " << GL.relevanter_Faktor;
      if (GL->relevanter_Faktor>=Primbasis_size) cerr << "zu grosser Wert!!" << endl;
      CGleichungszeile* pos = GLS[GL->relevanter_Faktor];
      //cerr << " ... ";
      if ( pos!=NULL )
	{ /* Relation mit diesem Eintrag existiert schon
	     -> neue Kombination mglich! */
	  //cerr << "kombiniere..." << endl;
          if ( GL->SizeOfRelation()<pos->SizeOfRelation() )
	    { /* weniger Primzahlen in der Relation bedeuten schnelleres Abarbeiten.
	  	 Da beide Gleichungszeilen quivalent sind, knnen sie problemlos so
		 getauscht werden, da die platzsparende im Gleichungssystem verbleibt. */
	      //clog << "Optimiere Relationen" << endl;
	      GLS[GL->relevanter_Faktor]=GL;
	      register CGleichungszeile* h = pos; pos=GL; GL=h; // Gleichungszeilen tauschen
	    }
 	  //GL->Ausgabe(); (*pos).Ausgabe();
 	  GL->kombinieren(*pos); 
	}
      else 
	{
	  GLS[GL->relevanter_Faktor]=GL; // Gleichungszeile ins Gleichungssystem eintragen
	  //GL->Relation_optimieren();
	  
	  /*
	    static int Optimierungszaehler = Primbasis_size/10;
	    if (--Optimierungszaehler==0)
	    {
	    Gleichungssystem_optimieren();
	    Optimierungszaehler=Primbasis_size/10;
	    }
	  */
	  // speichern
	  GL->speichern(StaticRelations_to_file);
	  Statusmeldung();
	  return;
	}
      //cerr << "ok." << endl;
    }
  //cerr << "Kombinierung abgeschlossen." << endl;
  
  /* Relation redundant	-> Faktorisierung probieren! */
  //cout << endl << "leere Relation entdeckt." << endl;
  //GL->Ausgabe();
  Statusmeldung();
  GL->Quadratische_Kongruenz_ausrechnen();
  //     GL->clear(); delete GL; // Gleichungszeile leeren
  delete GL;
#endif
}



TSiebElement log_Threshold = 0; // wird aber erst durch "Grundinitialisierung_des_Siebs" initialisiert

void SiebSchwellwert_kalkulieren()
{
  // Hier wird der Threshold (berschlagsmig) ermittelt
  
  // Voraussetungen zum Aufruf der Initialisierung:
  // 1. Logische Siebintervallgre M (fr Intervall [-M,M])
  // 2. kN (die mit Vorfaktor versehene zu faktorisierende Zahl
  // 3. die staische Faktorbasis
  // mssen bereits bekannt und berechnet sein.
 
  // logarithmischen Schwellwert kalkulieren:
  mpz_sqrt(x,kN); mpz_mul_ui(x,x,M); mpz_div_ui(x,x,20);
  double d = log(fabs(mpz_get_d(x))) - (Factor_Threshold*log(biggest_Prime_in_Factorbase));
  if (Primzahlen[1]==2) d-=log(2); // Falls 2 in der Faktorbasis, dann teilt 2 alle Siebeintrge
  log_Threshold=static_cast<TSiebElement>(log_Multiplikator*d);
}

void Sieb_initialisieren()
{
  //cout << "Starte Initialisierung des Siebs von " << Sieboffset
  //   << " bis " << Sieboffset+Siebgroesse << endl;

  /* log_Treshold bezeichnet den Schwellwert, der zum Erkennen der Relation
     unterschritten werden mu. Es ist besser, den Treshold bereits bei der
     Initialisierung zu bercksichtigen (d.h. abzuziehen), da beim Auslesen
     des Siebs dann nur auf negative Werte getestet werden braucht. */
  
  // nachfolgender Block dient (falls aktiviert) nur noch Validierungszwecken:
  /*
  Polynom.Werte_holen(Sieboffset,x,y);
  TSiebElement w_test = static_cast<TSiebElement>(log_Multiplikator*log(fabs(mpz_get_d(y))));
  TSiebElement FB_Threshold
    = static_cast<TSiebElement>(log_Multiplikator*Factor_Threshold*log(biggest_Prime_in_Factorbase));
  w_test-=FB_Threshold;
  if (Primzahlen[1]==2) w_test-=static_cast<TSiebElement>(log_Multiplikator*log(2));
  cout << "value " << y << " -> logwert: " << w_test << "<->" << log_Threshold << endl; 
  */

  register TSiebElement w = log_Threshold;
  /* log_Treshold bezeichnet den Schwellwert, der zum Erkennen der Relation
     unterschritten werden mu. Es ist besser, den Treshold bereits bei der
     Initialisierung zu bercksichtigen (d.h. abzuziehen), da beim Auslesen
     des Siebs dann nur auf negative Werte getestet werden braucht. */
  
  // Initialisierung des Siebs mit dem berschlagsmigen Schwellwert
  register int i=Siebgroesse; do Sieb[--i]=w; while (i);
  
  
  //cout << "Initialisierung abgeschlossen." << endl;
}


void schnelles_aussieben()
{
  //cout << "Aussieben mit Primzahlen der statischen FB" << endl;
     
  /* wir sieben erst ab einer greren Primzahl aus der Faktorbasis,
     da wir sowieso unsauber und schnell sieben...
     ( je kleiner die Primzahl, um langsamer ist das Sieben mit dieser
     Zahl; und bei der gewhlten Implementierung ohne Bercksichtigung
     von Primzahlpotenzen ist auch der zustzlich gemachte Fehler
     vernachlssigbar und wird zudem ber Factor_Threshold kompensiert.
     Aber Vorsicht: Shortcut-Konstante nicht zu hoch ansetzen, sonst fallen
     zu viele Relationen raus!
     Achtung: als Startwert mindestens 2 whlen, damit -1 und 2 ignoriert
              werden knnen. Dann gibt es auch 2 Deltawerte. )
  */
  for (int Primzahl_nr=sieben_ab_Primzahlindex; Primzahl_nr<Primbasis_size; Primzahl_nr++) 
    {
      register int Primzahl = Primzahlen[Primzahl_nr];
      register int d0=(Delta_der_Primzahlen[Primzahl_nr][0]-(Sieboffset+Siebgroesse))%Primzahl;
      register int d1=(Delta_der_Primzahlen[Primzahl_nr][1]-(Sieboffset+Siebgroesse))%Primzahl;

      register TSiebElement log_Primzahl = log_Primzahlen[Primzahl_nr];
      /* d.h. log_Primzahl erhlt den Wert (TSiebElement) (log_Multiplikator*log(Primzahl)) */
      
      //printf("Quadratisches Sieb mit %i\n",Primzahl);
      if (d1>=0) d1-=Primzahl;
      if (d0>=0) d0-=Primzahl;
      d0+=Siebgroesse;
      d1+=Siebgroesse;
      
      if (d0>d1) { int h=d0; d0=d1; d1=h; } // ggf. tauschen, so da d0<d1
      while (d0>=0)
       {
	 Sieb[d0]-=log_Primzahl; d0-=Primzahl;
	 Sieb[d1]-=log_Primzahl; d1-=Primzahl;
       }
      if (d1>=0) Sieb[d1]-=log_Primzahl;
    }
  
  for (int nr=0; nr<anz_weitere_Primzahlpotenzen; nr++)
    {
      register int PPotenz = Primzahlpotenzen[nr];
      register int d0=(Delta_der_Primzahlpotenzen[nr][0]-(Sieboffset+Siebgroesse))%PPotenz;
      register int d1=(Delta_der_Primzahlpotenzen[nr][1]-(Sieboffset+Siebgroesse))%PPotenz;
      
      register TSiebElement log_Primzahl = log_Primzahl_der_Primzahlpotenzen[nr];
      /* d.h. log_Primzahl erhlt den Wert (TSiebElement) (log_Multiplikator*log(Primzahl)) */
      
      // beachte: da mit den Primzahlen selbst auch gesiebt wird, darf nicht log(Primzahlpotenz), sondern
      // nur log(Primzahl) abgezogen werden!
      // Analogie: Wenn man schon durch p geteilt hat, mu man nocheinmal durch p teilen,
      //           um insgesamt durch p^2 zu teilen!
      
      if (d1>=0) d1-=PPotenz;
      if (d0>=0) d0-=PPotenz;
      d0+=Siebgroesse;
      d1+=Siebgroesse;
      
      if (d0>d1) { int h=d0; d0=d1; d1=h; } // ggf. tauschen, so da d0<d1
      while (d0>=0)
       {
	 Sieb[d0]-=log_Primzahl; d0-=PPotenz;
	 Sieb[d1]-=log_Primzahl; d1-=PPotenz;
       }
      if (d1>=0) Sieb[d1]-=log_Primzahl;      
    }

}


void aussieben_mit_Quadraten()
{
  // Quadrate von statischen (und ggf. dynamischen Faktoren) sieben.
  // Wichtig: Daher nur den einfachen Logarithmus des Faktors subtrahieren!
  // Bemerkung: Zur Effizienzsteigerung wird davon ausgegangen, da jedes Quadrat nur einmal
  //            innerhalb des gesamten Siebintervalls trifft!

  //cout << "Quadrate sieben Start" << endl;
  if (Sieb_Delta_QuadrateHeap.empty()) return; // es gibt nichts zu sieben...
  TSieb_Delta sieb_delta = Sieb_Delta_QuadrateHeap.top();
  register int delta = sieb_delta.delta-Sieboffset;
  
  while (delta<Siebgroesse)
    {
      Sieb_Delta_QuadrateHeap.pop(); // Element aus der Priorittswarteschlange nehmen
      Sieb[delta]-=static_cast<TSiebElement>(log_Multiplikator*log(sieb_delta.factor)); /* Teiler im Sieb vermerken */
      if (Sieb_Delta_QuadrateHeap.empty()) break;
      sieb_delta = Sieb_Delta_QuadrateHeap.top();
      delta = sieb_delta.delta-Sieboffset;
    }
  //cout << "Quadrate sieben Ende" << endl;
}


void aussieben_mit_Dynamic_Factoren()
{
  if (Sieb_Delta_Heap.empty()) return; // es gibt nichts zu sieben...
  short int HitCount=0;
  TSieb_Delta sieb_delta = Sieb_Delta_Heap.top();
  register int delta = sieb_delta.delta-Sieboffset;
  goto Schleifentest;
  
 Schleifenanfang_weiterer_Faktor:
  ++HitCount;
  //if (HitCount>=MaxHits) { cerr << "MaxHits erhhen!"; exit(1); } 
 Schleifenanfang_normal:
  Sieb_Delta_Heap.pop(); // Element aus der Priorittswarteschlange nehmen
  Sieb[delta]-=static_cast<TSiebElement>(log_Multiplikator*log(sieb_delta.factor)); /* Teiler im Sieb vermerken */
  
  // Da dynamische Faktoren selten auftreten, diese Faktoren schon whrend
  // des Siebens vermerken...
  // ...das erhht die Effizienz bei der Refaktorisierung...    
  Hits[HitCount].Faktor=sieb_delta.factor;
  
  sieb_delta.delta += sieb_delta.factor; // nchste Hit-Position         	      
  if (sieb_delta.delta<=M) // shortcut: nur dann eintragen, wenn die Relation noch im Siebintervall relevant wird
    Sieb_Delta_Heap.push(sieb_delta); // wieder in Priorittswarteschlange eintragen
  else
    if (Sieb_Delta_Heap.empty()) goto Schlange_leer; // keine weiteren Faktoren zum Sieben? 
  
  sieb_delta = Sieb_Delta_Heap.top(); // nchsten Eintrag aus Priorittswarteschlange holen
  register int delta2; delta2 = sieb_delta.delta-Sieboffset;
  if (delta==delta2) goto Schleifenanfang_weiterer_Faktor;  
  if (Sieb[delta]<0) // Relation gefunden?
    {
      Sieb[delta]=1; // weiteren Treffer an dieser Stelle abwrgen...
      Relation_einfuegen(new CGleichungszeile(delta+Sieboffset, HitCount+1)); /* dann einfgen */
    }
  delta=delta2;
  
  HitCount=0; // und auf ein neues
 Schleifentest:
  if (delta<Siebgroesse) goto Schleifenanfang_normal; // noch im aktuellen (physikalischen) Intervall?
  return; // fertig mit diesem Intervall
  
 Schlange_leer:
  if (Sieb[delta]<0) // Relation gefunden?
    {
      Sieb[delta]=1; // weiteren Treffer an dieser Stelle abwrgen...
      Relation_einfuegen(new CGleichungszeile(delta+Sieboffset,HitCount+1)); // dann einfgen
    }
  return; 
}


void schnelles_auslesen_des_Siebs()
{
  //cout << "Auslesen des (gesiebten) Siebintervalls" << endl;
  register int d=Siebgroesse;
  do
    if (Sieb[--d]<0) // Relation gefunden?
      Relation_einfuegen(new CGleichungszeile(Sieboffset+d)); /* dann einfgen */
  while (d);
}



void bestes_kN_ermitteln_mit_Knuth_Schroeppel()
{
  // es ist gnstig, mglichst viele kleine Primfaktoren in der Faktorbasis
  // zu haben. Hierfr kann man die zu faktorisierende Zahl mit einem kleinen
  // Vorfaktor (der am besten prim sein sollte) multiplizieren, um andere
  // Faktorbasen zu erhalten..
  // Welches der gnstigste Vorfaktor ist, wird hier ermittelt.
  // Da der Vorfaktor anschlieend nicht mehr bentigt wird, werden an dieser
  // Stelle die relavanten Werte kN = Vorfaktor*n sowie r = trunc(sqrt(kN))
  // berechnet.

  /* Bemerkung: Fr diese Implementierung des multipolynomialen Siebs ist es zustzlich
     erforderlich, da kN=1 (mod 4).
     Besser noch ist kN=1 (mod 8), wenn auch 2 in der Faktorbasis sein soll. (vgl. Paper von Silverman) */

  cout << "Ermittlung des Multipliers." << endl;
  
  const unsigned int k[] = {1,2,3,5,7,11,13,17,19,23,0};
  double bestw = -1e20; int best = 0;
  for (int i=0; k[i]!=0; i++)
    {
      mpz_mul_ui(kN,n,k[i]); /* kN initialisieren */
      if (mpz_mod_ui(x,kN,4)!=1) continue; // siehe Bemerkung

      cout << k[i] << ": ";
      
      double aktw = -0.5*log(k[i]);
      double g;
      unsigned int Primzahl = 1; // Standardmig ist 1+1 der erste Primzahlkandidat der FB 
      if (mpz_mod_ui(x,kN,8)!=1) Primzahl=2; // 2 nicht in der FB -> 2+1 erster Kandidat 
      for (int z=0; z<100; z++)
	{
	  do
	    {
	      do Primzahl++; while(!is_prime(Primzahl)); /* nchste Primzahl */
	      mpz_set_ui(x,Primzahl);
	    }
	  while (mpz_legendre(kN,x)==-1); // Primzahl gltig fr Primzahlbasis?
	  
	  // modified Knuth-Schroeppel-function:
	  if (Primzahl==2)
	    g=(mpz_mod_ui(x,kN,8)==1)? 2.0 : 0.0;
	  else 
	    g=(k[i]%Primzahl!=0)? 2/Primzahl : 1/Primzahl;
	  aktw+=g*log(Primzahl);
	  if (z<20) cout << Primzahl << " ";
	}
      cout << aktw << endl;
      if (aktw>bestw) { bestw=aktw; best=i; } // besserer Multiplikator...
    }
  cout << "Multiplikator fr Sieb ist " << k[best] << endl;
  Vorfaktor_von_kN=k[best]; // Vorfaktor global zugnglich machen
  mpz_mul_ui(kN,n,k[best]); /* kN initialisieren */
  mpz_sqrt(r,kN); /* Quadratwurzel von kN ermitteln */
  cout << r << " ist trunc(sqrt(kN))" << endl;
}


void bestes_kN_ermitteln()
{
  // es ist gnstig, mglichst viele kleine Primfaktoren in der Faktorbasis
  // zu haben. Hierfr kann man die zu faktorisierende Zahl mit einem kleinen
  // Vorfaktor (der am besten prim sein sollte) multiplizieren, um andere
  // Faktorbasen zu erhalten..
  // Welches ein gnstigster Vorfaktor ist, wird hier ermittelt.
  // Da der Vorfaktor anschlieend nicht mehr bentigt wird, werden an dieser
  // Stelle die relavanten Werte kN = Vorfaktor*n sowie r = trunc(sqrt(kN))
  // berechnet.

  /* Bemerkung: Fr diese Implementierung des multipolynomialen Siebs ist es zustzlich
     erforderlich, da kN=1 (mod 4).
     Besser noch ist kN=1 (mod 8), wenn auch 2 in der Faktorbasis sein soll.
     (vgl. Paper von Silverman) */

  // Hier nun eine besonders brutale Vorfaktorermittlung:
  const unsigned int P=19;
  // Es sollen alle Primzahlen von 2 bis P in der statischen Faktorbasis enthalten sein.
  // Dafr wird in Kauf genommen, da der Vorfaktor etwas grer wird...
  
  cout << "Ermittlung des Multipliers." << endl;
  
  unsigned int k=1; // testweiser Vorfaktor
  unsigned int p; // fr Primzahltest
  while (true)
    {
      mpz_mul_ui(kN,n,k); /* kN initialisieren */
      
      // ist 2 in der Faktorbasis?
      if (mpz_mod_ui(x,kN,8)!=1) goto next;
      
      cout << k << ", ";
      
      p=2;
      while (p<P)
	{
	  do p++; while(!is_prime(p));
	  mpz_set_ui(x,p); if (mpz_legendre(kN,x)==-1) goto next;
	}
      
      break; // // Vorfaktor geeignet!
      
    next: // nchsten Vorfaktor holen
      do k++; while(!is_prime(k));
    }
  cout << endl;
  
  Vorfaktor_von_kN=k; // Vorfaktor global zugnglich machen
  cout << "Multiplikator fr Sieb ist " << k << endl;
  mpz_sqrt(r,kN); /* Quadratwurzel von kN ermitteln */
  cout << r << " ist trunc(sqrt(kN))" << endl;
}


void CPolynom::erstes_Polynom_ermitteln(const mpz_t &fuer_kN, const int M)
{
  
  mpz_set(kN,fuer_kN);
  mpz_div_ui(kN_div2,kN,2); // kN/2 fr Berechnung negativer Modulowerte
  
  //mpz_t rkN; mpz_init(rkN); mpz_sqrt(rkN,kN);
  //const double W1=sqrt(2)/2;
  //const double W2=-1/(2*sqrt(2));
  //double A0=W1*mpz_get_d(rkN)/M;
  //double B0=0;
  //double C0=W2*M*mpz_get_d(rkN);
  //mpz_clear(rkN);
  //mpz_set_d(D,sqrt(A0)); // Initialwert fr Polynome

  // exakter:
  mpz_mul_ui(D,kN,2); mpz_sqrt(D,D); mpz_div_ui(D,D,2); mpz_div_ui(D,D,M); mpz_sqrt(D,D);
  // D ist der Initalwert fr das "optimale" Polynom
  
  // nun zur eigentlichen Polynomberechnung:
  // Let A ~ D^2, D (probab. prime), (D/kN)=1, D=3 (mod 4)
  while (mpz_mod_ui(x,D,4)!=3) mpz_add_ui(D,D,1);
  naechstes_Polynom_ermitteln(); // hier folgt die weitere Berechnung
}

void CPolynom::naechstes_Polynom_ermitteln(const int step=0)
{
  mpz_add_ui(D,D,step*4);
  // Let A ~ D^2, D (probab. prime), (D/kN)=1, D=3 (mod 4)
  do mpz_add_ui(D,D,4); while ( (!mpz_probab_prime_p(D,25)) || (mpz_jacobi(D,kN)!=1) );
  mpz_mul(A,D,D); mpz_mul_ui(A2,A,2);
  
  // h0 = (kN)^((D-3)/4) mod D
  mpz_sub_ui(h0,D,3); mpz_div_ui(h0,h0,4); mpz_powm(h0,kN,h0,D);
  
  // h1 = kN*h0 mod D
  mpz_mul(h1,kN,h0); mpz_mod(h1,h1,D);
  //check: h1=(kN)^((D+1)/4) mod D
  mpz_add_ui(x,D,1); mpz_div_ui(x,x,4); mpz_powm(x,kN,x,D);
  if (mpz_cmp(h1,x)!=0) { cerr << "Berechnungsfehler: h1 ist falsch!" << endl; exit(1); }
  
  // h2 = (2*h1)^-1 * ((kN-h1^2) /D) mod D
  mpz_mul(h2,h1,h1); mpz_sub(h2,kN,h2);
  {
    mpz_mod(x,h2,D);
    if (mpz_cmp_ui(x,0)!=0)
      { 
	cerr << "Fehler: Division mit Rest ";
	mpz_out_str(stderr,10,x); cerr << " !" << endl;
	exit(1);
      }
  }
  mpz_div(h2,h2,D);
  mpz_mul_ui(x,h1,2); mpz_invert(x,x,D); mpz_mul(h2,x,h2);
  mpz_mod(h2,h2,D);
  
  // B = h1 + h2*D mod A
  mpz_mul(B,h2,D); mpz_add(B,B,h1); mpz_mod(B,B,A);
  if (mpz_mod_ui(x,B,2)==0) mpz_sub(B,A,B); 
  
  // C = (B^2-kN) / (4*A)
  mpz_mul(C,B,B); mpz_sub(C,C,kN); mpz_div(C,C,A); mpz_div_ui(C,C,4);
  
  cout << "Die Parameter des Polynoms A*x^2+B*x+C lauten: " << endl;
  cout << "A=" << A << endl;
  cout << "B=" << B << endl;
  cout << "C=" << C << endl;
  cout << "D=" << D << " mit A=D^2" << endl;
  
  mpz_mul_ui(D2_inv_mod_kN,D,2); mpz_invert(D2_inv_mod_kN,D2_inv_mod_kN,kN);
  //testen();
}

void CPolynom::testen(const signed int Siebpos=0)
{
  mpz_set_si(h2,Siebpos);
  
  // Variante 1: ( (2*A*wert+B) / (2*D) )^2 mod kN
  mpz_mul_ui(x,A,2); mpz_mul(x,x,h2); mpz_add(x,x,B);
  //mpz_div(x,x,D); mpz_div_ui(x,x,2);
  mpz_mul(x,x,D2_inv_mod_kN); mpz_mod(x,x,kN);
  mpz_mul(x,x,x); mpz_mod(x,x,kN);
  if (mpz_cmp(x,kN_div2)>0) { mpz_sub(x,kN,x); mpz_neg(x,x); } // ggf. negativen Wert berechnen!
  
  // Variante 2: A*wert^2 + B*wert + C mod kN
  
  mpz_mul(h0,A,h2); mpz_mul(h0,h0,h2);
  mpz_mul(h1,B,h2);
  mpz_add(y,h0,h1); mpz_add(y,y,C); mpz_mod(y,y,kN);
  if (mpz_cmp(y,kN_div2)>0) { mpz_sub(y,kN,y); mpz_neg(y,y); } // ggf. negativen Wert berechnen!
  
  cout << "Test: Q(" << Siebpos << ")" << endl;
  cout << " =" << x << endl;
  cout << " =" << y << endl;

  if (mpz_cmp(x,y)!=0)
    {
      cerr << "Fehler: Werte sind verschieden!" << endl;
      exit(1);
    }
}

void CPolynom::Werte_holen(const signed int Siebpos, mpz_t &Wurzelwert, mpz_t &Q) const
{  // Variante 1: ( (2*A*wert+B) / (2*D) )^2 mod kN
  mpz_set_si(Wurzelwert,Siebpos);
  mpz_mul(Wurzelwert,Wurzelwert,A2);
  mpz_add(Wurzelwert,Wurzelwert,B);
  mpz_mul(Wurzelwert,Wurzelwert,D2_inv_mod_kN); mpz_mod(Wurzelwert,Wurzelwert,kN);
  
  mpz_mul(Q,Wurzelwert,Wurzelwert); mpz_mod(Q,Q,kN);
  if (mpz_cmp(Q,kN_div2)>0) mpz_sub(Q,Q,kN); // ggf. negativen Wert berechnen!
  
  //cout << "Werte fr " << Siebpos << " ";
  //mpz_out_str(stdout,10,Wurzelwert); cout << " -> ";
  //mpz_out_str(stdout,10,Q); cout << endl;
}


void CPolynom::Daten_sichern(ostream &ostr) // fr Recovery etc.
{
  // Daten sichern, die fr die Ermittlung des aktuellen Polynoms wichtig sind
  // D ist der Wert, der das aktuelle Polynom aus der Funktionsschar bestimmt.
  ostr << D << endl;
}

void CPolynom::Daten_laden(istream &in) // fr Recovery etc.
{
  // Daten laden, die fr die Ermittlung des aktuellen Polynoms wichtig sind 
  in >> D; in.ignore(1,'\n');
  mpz_sub_ui(D,D,4); // um vier vermindern
  naechstes_Polynom_ermitteln(); // und neu aufsetzen
}

void alle_Deltas_neu_berechnen()
{
  /* Beim Wechsel auf ein neues Polynom werden alle bisherigen Deltawerte der Primzahlen
     ungltig. Daher ist es erforderlich, diese neu zu berechnen */

  cout << "Berechne die Deltas neu" << endl;

  // zunchst alle (mglicherweise) noch bestehenden dynamischen Faktoren aus der Warteschlange nehmen:
  //Sieb_Delta_Heap.clear();
  while (!Sieb_Delta_Heap.empty()) Sieb_Delta_Heap.pop(); // fuer priority_queue, die kein clear hat

  // ebenso die Quadrate:
  //Sieb_Delta_QuadrateHeap.clear();
  while (!Sieb_Delta_QuadrateHeap.empty()) Sieb_Delta_QuadrateHeap.pop(); // fuer priority_queue, die kein clear hat
 
  
  // als erstes die Deltas aus der statische Faktorbasis ermitteln
  mpz_t inv_A2, P, wuq;
  mpz_init(inv_A2); mpz_init(P); mpz_init(wuq);
  int i=1;
  if (Primzahlen[i]==2) i++; // Primzahl 2 berspringen, da Polynom damit Probleme hat...
  for (; i<Primbasis_size; i++)
    {
      mpz_set_ui(P,Primzahlen[i]);
      mpz_invert(inv_A2,Polynom.A2,P);
      
      mpz_set_ui(x,SQRT_kN_der_Primzahlen[i]);
      mpz_sub(x,x,Polynom.B); mpz_mul(x,x,inv_A2); 
      Delta_der_Primzahlen[i][0]=mpz_mod_ui(x,x,Primzahlen[i]);
      
      mpz_set_ui(x,SQRT_kN_der_Primzahlen[i]); mpz_neg(x,x);
      mpz_sub(x,x,Polynom.B); mpz_mul(x,x,inv_A2); 
      Delta_der_Primzahlen[i][1]=mpz_mod_ui(x,x,Primzahlen[i]);

      //ggf. testen, ob Werte in Ordnung sind...
      /*
      Polynom.Werte_holen(Delta_der_Primzahlen[i][0],x,y);
      if (mpz_mod_ui(y,y,Primzahlen[i])!=0) 
	{ 
	  cerr << "Deltawert0 " << Delta_der_Primzahlen[i][0] << " liefert Teiler fr "
	       << Primzahlen[i] << " nicht! (Rest ";
	  mpz_out_str(stdout,10,y); cout << ")" << endl;
	  Delta_der_Primzahlen[i][0]-=mpz_get_ui(y);
	  if (Delta_der_Primzahlen[i][0]<0) Delta_der_Primzahlen[i][0]+=Primzahlen[i];
	}
      Polynom.Werte_holen(Delta_der_Primzahlen[i][1],x,y);
      if (mpz_mod_ui(y,y,Primzahlen[i])!=0)
	{ 
	  cerr << "Deltawert1 " << Delta_der_Primzahlen[i][1] << " liefert Teiler fr "
	       << Primzahlen[i] << " nicht! (Rest ";
	  mpz_out_str(stdout,10,y); cout << ")" << endl;
	  Delta_der_Primzahlen[i][1]-=mpz_get_ui(y);
	  if (Delta_der_Primzahlen[i][1]<0) Delta_der_Primzahlen[i][1]+=Primzahlen[i];
	}
      */

      if (i>FB_maxQuadrate && Vorfaktor_von_kN%Primzahlen[i]!=0)
	{
	  // Fr alle "groen" Quadrate aus der FB auch die Deltas ermitteln,
	  // aber diese (falls innerhalb des Siebintervalls liegend) zusammen mit den dynamischen Faktoren sieben.
	  mpz_set_ui(P,Primzahlen[i]); mpz_mul(P,P,P); // Quadrat der Primzahl holen
	  mpz_invert(inv_A2,Polynom.A2,P);

	  // nun die Wurzel modulo P ermitteln (auf ein Caching wird verzichtet, da sie gro sein kann)
	  // Wurzeln von kN modulo Primzahl^2 ermitteln
	  int Wu = SQRT_kN_der_Primzahlen[i];
	  mpz_set_ui(y,2*Wu);
	  if (mpz_invert(y,y,P)==0)
	    {
	      cerr << "PQ-Wurzel fr " << Primzahlen[i] << ": Inverses existiert nicht!" << endl;
	      exit(1);
	    }
	  
	  mpz_mod(x,kN,P); mpz_set_ui(wuq,Wu); mpz_mul_ui(wuq,wuq,Wu); mpz_sub(x,x,wuq);
	  mpz_mul(x,x,y); mpz_add_ui(x,x,Wu); mpz_mod(wuq,x,P);
	  
	  // ggf. Prfung, ob Wurzel korrekt ist:
	  /* 
	     mpz_mul(x,wuq,wuq); mpz_sub(x,kN,x); mpz_mod(x,x,P);
	     if (mpz_cmp_ui(x,0)!=0)
	      {
	        cerr << "PQ-Wurzel nicht korrekt!" << endl;
	        cerr << "Primzahl=" << Primzahlen[i] << endl;
	        cerr << "kN= " << kN << endl;
	        cerr << "SQRT(kN) mod p=" << SQRT_kN_der_Primzahlen[i] << endl;
	        cerr << "SQRT(kN) mod p^2=" << wuq << endl;
	        cerr << "aber falsch: " << x << endl;
	        exit(1); 
	      }
	  */
	  
	  // nun die Deltas:

	  // Delta1
	  mpz_sub(x,wuq,Polynom.B); mpz_mul(x,x,inv_A2);mpz_mod(x,x,P);
	  mpz_sub_ui(x,x,M);
	  if (mpz_cmp_ui(x,0)<0) mpz_add_ui(x,x,M);
	  if (mpz_cmp_ui(x,M)<=0) // liegt der Wert im Intervall?
	    {
	      TSieb_Delta sieb_delta;
	      sieb_delta.factor=Primzahlen[i];
	      sieb_delta.delta=mpz_get_ui(x);
	      Sieb_Delta_QuadrateHeap.push(sieb_delta); // und ablegen
	      //cout << "Quadrat von " << sieb_delta.factor << " gepusht!" << endl;
	    }
	  
	  // Delta2
	  mpz_neg(x,wuq); mpz_sub(x,x,Polynom.B); mpz_mul(x,x,inv_A2); mpz_mod(x,x,P);
	  mpz_sub_ui(x,x,M);
	  if (mpz_cmp_ui(x,0)<0) mpz_add_ui(x,x,M);
	  if (mpz_cmp_ui(x,M)<=0) // liegt der Wert im Intervall?
	    {
	      TSieb_Delta sieb_delta;
	      sieb_delta.factor=Primzahlen[i];
	      sieb_delta.delta=mpz_get_ui(x);
	      Sieb_Delta_QuadrateHeap.push(sieb_delta); // und ablegen
	      //cout << "Quadrat von " << sieb_delta.factor << " gepusht!" << endl;
	    }
	  
	}
      
    }
  cout << "Es wurden " << Sieb_Delta_QuadrateHeap.size() << " Quadrate gepusht." << endl;
  
  for (i=0; i<anz_weitere_Primzahlpotenzen; i++)
    {
      // Fr diese Zahlen auch die Deltas bzgl. der jeweiligen Potenz ermitteln
      // (fr das Sieben zusammen mit den statischen Faktoren)
      mpz_set_ui(P,Primzahlpotenzen[i]);
      mpz_invert(inv_A2,Polynom.A2,P);
      
      mpz_set_ui(x,SQRT_kN_der_Primzahlpotenzen[i]);
      mpz_sub(x,x,Polynom.B); mpz_mul(x,x,inv_A2); 
      Delta_der_Primzahlpotenzen[i][0]=mpz_mod_ui(x,x,Primzahlpotenzen[i]);
      
      mpz_set_ui(x,SQRT_kN_der_Primzahlpotenzen[i]); mpz_neg(x,x);
      mpz_sub(x,x,Polynom.B); mpz_mul(x,x,inv_A2); 
      Delta_der_Primzahlpotenzen[i][1]=mpz_mod_ui(x,x,Primzahlpotenzen[i]);
    }
  
  mpz_clear(inv_A2); mpz_clear(P); mpz_clear(wuq);
  
  // dann die dynamische Faktorbasis behandeln:
  TDynamic_Factor_Relations::iterator pos;
  for (pos=Dynamic_Factor_Relations.begin();
       pos!=Dynamic_Factor_Relations.end(); ++pos) Deltas_berechnen_fuer(*pos, -M);
  
  cout << "Deltaberechnung beendet." << endl;
}


bool get_Zahl (mpz_t &n, char* &s, char op='R')
{ 
  /* Zeichenkette in multiple-precision-Zahl umwandeln.
     Die Zeichenkette kann ein beliebiger Term mit Klammern,
     und den Operationszeichen +,-,*,^ sein, solange er numerisch
     auswertbar ist. */
  
  while (s[0]==' ') s++; // ggf. Leerzeichen berlesen
  
  bool term_structure_ok = true;
  char* s_anf = s; // Start-Position des aktuellen Subterms
  char h;
  mpz_t tmp;
  mpz_init(tmp);
  
  if (s[0]=='(')
    {
      s++; // ffnende Klammer berspringen
      term_structure_ok=get_Zahl(tmp,s); // geklammerten Term holen
      if (s[0]==')') s++; // schlieende Klammer berspringen
      else 
	{
	  term_structure_ok=false;
	  cerr << "error: ')' expected at '" << s << "'" << endl; 
	}
    }
  else
    { // ungeklammerten Term holen
      char* anf=s;
      while (s[0]>='0' && s[0]<='9') s++;
      h=s[0]; s[0]='\0';
      if (mpz_set_str(tmp, anf, 10))
	{
	  term_structure_ok=false;
	  s[0]=h;
	  cerr << "error: number expected at '" << anf << "'" << endl;
	}
      s[0]=h;
    }
  
  while (s[0]==' ') s++; // Leerzeichen berlesen
  h=s[0]; // nchstes Operationszeichen
  
  if (term_structure_ok) switch (op)
    {
    case 'R' :
      //cout << "Reset" << endl;
      mpz_set(n,tmp);
      break;   
    case '+' :
      if (h=='*' || h=='^')
	{
	  //cout << "Punkt vor Strich" << endl; 
	  term_structure_ok=get_Zahl(tmp,++s,h); h='\0';
	}	
      //cout << "Addition" << endl;	
      mpz_add(n,n,tmp);
      break;
    case '-' :
      if (h=='*' || h=='^')
	{
	  //cout << "Punkt vor Strich" << endl; 
	  term_structure_ok=get_Zahl(tmp,++s,h); h='\0';
	}	
      //cout << "Subtraktion" << endl;
      mpz_sub(n,n,tmp);
      break;
    case '*' : 
      if (h=='^')
	{
	  //cout << "Exponentiation vor Multiplikation" << endl; 
	  term_structure_ok=get_Zahl(tmp,++s,h); h='\0';
	}	
      //cout << "Multiplikation" << endl;
      mpz_mul(n,n,tmp);
      break;
    case '^' :
      if (h=='^') cout << "Warnung: a^b^c ambiguent. Setze (a^b)^c." << endl; 
      // cout << "Exponentiation" << endl;
      mpz_pow_ui(n,n,mpz_get_ui(tmp));
      break;
    default :
      s=--s_anf;
      cerr << "error: operation +-*^ expected at '" << s << "'" << endl;	
      term_structure_ok=false;
      break;	 
    }
  mpz_clear(tmp);
  if (term_structure_ok && h!='\0' && h!=')' ) return get_Zahl(n,++s,h);
  else return term_structure_ok;
}

#ifdef IS_SERVER

void bearbeite_daten_stream(iostream &daten)
{
  string s;
  Tfactor factor;
  TDynamic_Factor_Relations::iterator pos;
  CGleichungszeile* GL;
  daten >> s;
  
  if (s == "Hunger!")
    {
      cout << "Sende zu faktorisierende Zahl." << endl;
      daten << n << endl;
      return;
    }
  
  if (s=="DynamicFactors?_ab_fpos")
    {
      streampos startfpos; daten >> startfpos;
      cout << "Sende dynamische Faktoren ab Fileposition " << startfpos << endl;
      DynamicRelations_to_file.seekp(0,ios::end); daten << DynamicRelations_to_file.tellp() << " ";
      for (pos= Dynamic_Factor_Relations.begin();
	   pos != Dynamic_Factor_Relations.end(); ++pos)
	if ((*pos).fpos>=startfpos)
	  {
	    daten << (*pos).factor << " ";
	  }
      daten << "-1" << endl;
      return;
    }
  
  if (s=="Polynom?")
    {
      int intervall;
      daten >> intervall;
      cout << "Sende neues Polynomintervall (" << intervall << ")" << endl;
      Polynom.Daten_sichern(daten);
      Polynom.naechstes_Polynom_ermitteln(intervall);
      Polynom.Daten_sichern(daten);
      
      // fr Recovery:
      streampos fpos = Recovery_to_file.tellp();
      Polynom.Daten_sichern(Recovery_to_file); // aktuelles Polynom abspeichern
      Recovery_to_file.seekp(fpos);
      return;
    }
  
  if (s == "Relation!")
    {
      static int counter = 100;
      if (--counter==0) { counter=100; Statusmeldung(); }
      //cout << "Empfange gefundene Relation." << endl;
  
      GL = new CGleichungszeile();
      factor = GL->kombinieren(daten);
      if (factor == 1)
	{
	  // cout << "Habe statische Relation empfangen!" << endl;
	  gefundene_Relationen++;
	  Relation_einfuegen(GL);
	  return;
	}
      mpz_set_d(x, factor);
      if (mpz_probab_prime_p(x, 25))
	{
	  // Dynamic Factor
	  TDynamicFactorRelation relation;
	  relation.factor = static_cast<int>(factor);
	  pos = Dynamic_Factor_Relations.find(relation); // dynamischen Faktor suchen
	  if (pos == Dynamic_Factor_Relations.end())
	    { // noch nicht bekannt
	      relation.fpos = GL->speichern(DynamicRelations_to_file,relation.factor);
	      relation.SQRT_kN=SQRT_kN_mod_Primzahl(relation.factor);
	      if (relation.factor<Dynamicfactor_Cache_Treshold)
		{
		  relation.GL=GL;
		}
	      else
		{
		  relation.GL=NULL;
		  delete GL;
		}
	      Dynamic_Factor_Relations.insert(relation);
	      Special_Factoren_durch_Primfaktor_splitten(relation.factor);
	    } 
	  else
	    { // schon bekannt
	      //cout << "Kann dynamische Relation zu statischer kombinieren!" << endl;
	      mpz_mul(GL->Delta,GL->Delta,x); mpz_mod(GL->Delta,GL->Delta,n);
	      GL->kombinieren(DynamicRelations_from_file, (*pos).fpos);
	      gefundene_Relationen++;
	      Relation_einfuegen(GL);
	    }
	  return;
	}
      // Special-Factor
      // Specialfactor mit Dynamicfactoren testen:
      /* Anmerkung:
	 Beim verteilten Sieben kann es passieren, da ein (notwendigerweise zusammengesetzter)
	 Special-Factor von einem Client entdeckt wird, der nicht mit dem aktuellen Satz der
	 dynamischen Faktoren siebt. - In diesem Fall, der mit steigender Anzahl der Clients immer
	 wahrscheinlicher wird, knnte es sinnvoll sein, Specialfaktoren auf ihre Teilbarkeit
	 mit den vorliegenden dynamischen Faktoren des Servers zu testen und ggf. zu splitten.
	 Allerdings bringt ein Verzicht auf diesen Test keine weiteren Nachteile mit sich.
	 Wir verzichten daher auf diese - nur beim verteilten Sieben mgliche - Prozedur.
      */
      // if(!Special_Factor_durch_Dynamicfactoren_splitten(factor)) // nicht implementiert (s.Anm.)
      {
	// Echter Specialfactor
	TSpecial_Factor_Relations::iterator pos;
	TSpecialFactorRelation relation;
	relation.factor = factor;
	pos = Special_Factor_Relations.find(relation);
	if (pos==Special_Factor_Relations.end())
	  {
	    relation.fpos=GL->speichern(SpecialRelations_to_file,relation.factor);
	    Special_Factor_Relations.insert(relation);
	    delete GL; // und Speicher der Relation wieder freigeben
	  }
	else
	  {
	    // Special-Factor ist bereits bekannt
	    Special_hit++;
	    gefundene_Relationen++;
	    mpz_mul(GL->Delta,GL->Delta,x); mpz_mod(GL->Delta,GL->Delta,n);
	    GL->kombinieren(SpecialRelations_from_file,(*pos).fpos);
	    Relation_einfuegen(GL);
	  }
      }
      return;
    }
  
  cerr << "Falsche Anfrage \"" << s << "\"" << endl;
}

#endif


int main(int argc, char* argv[])
{
  bool recover=false;

  cout << " --------------------------------------" << endl;
  cout << "-     Faktorisierungsprogramm V2.04     -" << endl;
  cout << "-  written 11/1998 by Thorsten Reinecke -" << endl;
  cout << "-     (reinecke@fmi.uni-passau.de)      -" << endl;
  cout << " --------------------------------------" << endl;
  cout << endl;
  cout << "This program may be freely distributed under the" << endl;
  cout << "terms of the GNU-Public-Licence." << endl;
  cout << endl;
    
  if (argc>2)
    {
      cerr << "Zu faktorisierende Zahl als Parameter erwartet!" << endl;
      exit(0);
    }
  
  recover = (argc==1); // ohne Argument: Recovery-Modus 

  mpz_init(n); /* Variable fr zu faktorisierende Zahl bereitstellen */
  mpz_init(kN);
  mpz_init(r);
  mpz_init(x);
  mpz_init(y);

#ifndef IS_CLIENT
  if (!recover) // Nicht im Recovery-Modus -> Argumente auswerten
    { 
      // bergebene Zahl holen (und ggf. ausrechnen)
      char* neuer_str = strdup(argv[1]);
      char* str = neuer_str; 
      if (!get_Zahl(n, str)) 
	{
	  printf("Falsche Eingabe ab: '%s'\n", str);
	  exit(0);
	}
      else
	if (str[0]!='\0')
	  {    
	    printf("Klammerungsstruktur falsch!\n");
	    exit(0);
	  }
	else
	  if (mpz_cmp_ui(n,0)<=0)
	    {
	      printf("Eingegebene Zahl sollte positiv sein!\n");
	      exit(0);
	    }
      delete [] neuer_str;
    }

  if (recover)
    {
      // Recovery: Fortsetzung einer abgebrochenen Faktorisierung
      // Streams ffnen
      StaticRelations_buffer.open(StaticRelationsFile,ios::in|ios::out);
      SpecialRelations_buffer.open(SpecialRelationsFile ,ios::in|ios::out);
      DynamicRelations_buffer.open(DynamicRelationsFile,ios::in|ios::out);
      Recovery_buffer.open(RecoveryFile,ios::in|ios::out|ios::ate);
      
      if ( (!Recovery_from_file) || (Recovery_from_file.tellg()<1) )
	{
	  cerr << "Recovery nicht mglich!" << endl;
	  exit(1);
	}
      
      Recovery_from_file.seekg(0,ios::beg);
      Recovery_to_file.seekp(0,ios::beg);
      Recovery_from_file >> n; // Zahl holen
      Recovery_from_file.ignore(1,'\n');
      cout << "Setzte Faktorisierung mit Quadratischem Sieb fort fr" << endl;
      cout << n << endl; 
    }
  else
    {
      // Neustart: Starten einer neuen Faktorisierung
      // Streams ffnen
      StaticRelations_buffer.open(StaticRelationsFile,ios::out|ios::trunc);
      SpecialRelations_buffer.open(SpecialRelationsFile ,ios::in|ios::out|ios::trunc);
      DynamicRelations_buffer.open(DynamicRelationsFile,ios::in|ios::out|ios::trunc);
      Recovery_buffer.open(RecoveryFile,ios::in|ios::out|ios::trunc);
      
      cout << "Starte Faktorisierung fr" << endl;
      cout << n << endl;
      
      Factorization_to_file << endl << "Factorization of " << argv[1] << ":" << endl;
      Factorization_to_file << n << " = " << endl;
      
      easy_factor(); // "einfache" Faktoren rausstreichen
      
      cout << "Starte Faktorisierung mit Quadratischem Sieb fr" << endl;
      cout << n << endl;
      Recovery_to_file << n << endl;           
    }
#else
  if (recover)
    {
      cerr << "Fehlender Parameter" << endl;
      exit(1);
    }
  communication_name = argv[1];
#ifdef USE_NETWORK
  // existiert der rechner?
  struct hostent *hp;
  hp = gethostbyname(communication_name);
  if (hp == NULL)
    {
      cerr << "Unbekannter Rechner " << communication_name << endl;
      exit (1);
    }
  
  {
    unix_io_client communication_stream(communication_name, server_port);
    communication_stream << "Hunger!" << endl;
    communication_stream >> n;
    cout << "n=" << n << endl;
  }
#else
  // existiert das file?
  ifstream communication_stream(communication_name);
  if (!communication_stream)
    {
      cerr << "Kann File " << communication_name << " nicht oeffnen" << endl;
      exit (1);
    }
  char *tmp = communication_name;
  communication_name = new char [strlen(tmp)+5];
  strcpy(communication_name, tmp);
  strcat(communication_name, ".out");
  
  communication_stream >> n; communication_stream.ignore(1,'\n');
  cout << "n=" << n << endl;
#endif
#endif  
  
  bestes_kN_ermitteln(); /* optimaler Wert, fr den gesiebt werden kann */
  tune_parameters(mpz_sizeinbase(kN,10)); // Parametertuning
  
  if (mpz_cmp_ui(r,Siebgroesse)<0)
    {
      printf("Sieb ist berdimensioniert (ggf. verkleinern)!\n");
      exit(1);
    }
  
  Statische_Faktorbasis_ermitteln();
  cout << "Die ersten 20 Elemente der statischen FB lauten" << endl;
  for (int i=0; i<20; i++) cout << Primzahlen[i] << " "; cout << endl;
  cout << "Grte Primzahl in statischer Faktorbasis: " << biggest_Prime_in_Factorbase << endl;
  cout << "Anzahl Primzahlquadrate in FB: " << FB_maxQuadrate << endl;
  cout << "Anzahl Primzahlpotenzen in FB: " << anz_weitere_Primzahlpotenzen << endl;

  SiebSchwellwert_kalkulieren(); // Schwellwert zur Erkennung der Relationen fr Siebvorgang bestimmen
  Sieb=new TSiebElement [Siebgroesse]; /* Speicher fr das Sieb allokieren */
  
#ifndef IS_CLIENT
  if (recover)
    {
      // Static Relations einlesen
      StaticRelations_from_file.seekg(0,ios::beg);
      while (StaticRelations_from_file.peek()=='G') 
	{
	  CGleichungszeile* GL = new CGleichungszeile();
	  GL->kombinieren(StaticRelations_from_file);
	  GLS[GL->relevanter_Faktor]=GL; gefundene_Relationen++;
	  StaticRelations_from_file.ignore(1,'\n');
	}
      StaticRelations_from_file.clear();
      cout << "Recover: " << gefundene_Relationen << " Static Relations eingelesen" << endl;

      // Dynamic Relations einlesen
      DynamicRelations_from_file.seekg(0,ios::beg);
      while (DynamicRelations_from_file.peek()=='G') 
	{
	  TDynamicFactorRelation relation;
	  relation.fpos = DynamicRelations_from_file.tellg();
	  string s; DynamicRelations_from_file >> s;
	  DynamicRelations_from_file >> relation.factor;
	  relation.SQRT_kN=SQRT_kN_mod_Primzahl(relation.factor);
	  relation.GL=NULL;
	  if (relation.factor<Dynamicfactor_Cache_Treshold)
	    {
	      relation.GL = new CGleichungszeile();
	      relation.GL->kombinieren(DynamicRelations_from_file,relation.fpos); // GL einlesen
	    }
	  DynamicRelations_from_file.ignore(1000000,'\n');
	  Dynamic_Factor_Relations.insert(relation);
	}
      DynamicRelations_from_file.clear();
      cout << "Recover: " << Dynamic_Factor_Relations.size()
	   << " Dynamic Relations eingelesen" << endl;

      // Special Relations einlesen
      SpecialRelations_from_file.seekg(0,ios::beg);
      while (true) 
	{
	  while (SpecialRelations_from_file.peek()=='U')
	    SpecialRelations_from_file.ignore(1000000,'\n'); // ungltige Zeilen berlesen
	  if (SpecialRelations_from_file.peek()!='G') break;
	  TSpecialFactorRelation relation;
	  relation.fpos = SpecialRelations_from_file.tellg();
	  string s; SpecialRelations_from_file >> s;
	  SpecialRelations_from_file >> relation.factor;
	  Special_Factor_Relations.insert(relation);
	  SpecialRelations_from_file.ignore(1000000,'\n');
	}
      SpecialRelations_from_file.clear();
      cout << "Recover: " << Special_Factor_Relations.size()
	   << " Special Relations eingelesen" << endl;    
      
      
#ifdef IS_SERVER
      // beim verteilten Sieben knnten den Clients einige potentielle dynamische Faktoren
      // durch die Lappen gegangen sein. Dies geschieht, falls ein Client einen Special-Factor
      // findet, der nicht mit dem vollen Satz der aktuellen (und daher nur dem Server bekannten)
      // dynamischen Faktoren gesiebt wurde.
      // Anstelle eines Tests zur Laufzeit werden diese im Normalfall eher seltenen Flle
      // nun hier beim Neu-Hochfahren des Servers bercksichtigt:
      // WICHTIG: erst mssen alle dynamischen Faktoren eingelesen sein! Ansonsten knnen
      //          doppelt auftretende dynamische Faktoren nicht zu einer Relation reduziert werden!
      // folgende Anweisungen wrden das bewerkstelligen, aber es lohnt sich kaum und dauert zu lange:
      // (-> irgendwann wird die Implementation eh auf faktorisierte Special-Faktoren umgestellt)

      //for (TDynamic_Factor_Relations::iterator pos = Dynamic_Factor_Relations.begin();
      //   pos!=Dynamic_Factor_Relations.end(); ++pos)
      //	Special_Factoren_durch_Primfaktor_splitten((*pos).factor);
#endif
      
    }
#else // is client:
#ifndef USE_NETWORK 
  TDynamicFactorRelation relation;
  relation.fpos = 0;
  communication_stream >> relation.factor; // filepos ueberlesen
  while (true)
    {
      communication_stream >> relation.factor;
      // cout << "*" << relation.factor << "*" << flush;
      if (relation.factor<=0) break;
      relation.SQRT_kN=SQRT_kN_mod_Primzahl(relation.factor);
      Dynamic_Factor_Relations.insert(relation);
    }
  //cout << endl;
#endif
#endif  
  
  
  Polynom.erstes_Polynom_ermitteln(kN,M); // erstes Polynom ermitteln
  
#ifndef IS_CLIENT
  if (recover)
    {
      streampos fpos = Recovery_from_file.tellg();
      Polynom.Daten_laden(Recovery_from_file); // aktuelles Polynom (zum Fortsetzen) ermitteln
      Recovery_to_file.seekp(fpos);
    }
#endif


  cout << "#Relationen{Verteilung}/#Special-Factors(Verteilung)/Faktorbasis/dynamicFB"
       << " [Siebpos] -> Fortschritt" << endl;


#ifdef IS_SERVER
  int server_socket = open_internet_port(server_port,10);
  if (server_socket<0) {
    cerr << "Kann Port " << server_port << " nicht oeffnen" << endl;
    exit(1);
  }
  
  cout << "Waiting for clients..." << endl;

  while (true)
    {    
      
      int communication_socket = accept(server_socket, NULL, NULL);
      if (communication_socket > 0)
	{
	  unix_buffer buffer(communication_socket);
	  iostream communication_stream(&buffer);
	  
	  bearbeite_daten_stream(communication_stream);
	  communication_stream << flush;
	  close (communication_socket);
	}
      
      // collecting_phase_beendet= ( gefundene_Relationen > 0.6*Primbasis_size );
      /* Durch das Flag "collecting_phase_beendet" ist es mglich, das Aufsammeln weiterer dynamischer
	 oder special-Faktoren zu unterbinden. Dies knnte ratsam sein, wenn der Haupt- oder auch
	 der Hintergrundspeicher knapp wird oder das Programm schon solche Fortschritte im
	 Sammeln der eigentlichen Relationen macht, da dynamische oder special-Faktoren hinderlich
	 wirken.
	 Da aber eine Speicherkanppheit bei UNIX nicht sinnvoll ermittelt werden kann (andere Prozesse
	 knnen jederzeit Speicher anfordern/freigeben), empfiehlt es sich eher, das Programm
	 bei Speicherknappheit abzubrechen, die dynamic_relations-Datei zu krzen/zu lschen und
	 das Programm anschlieend neu zu starten. Auf diese Weise kann man die Speichernutzung
	 des Programms wesentlich flexibler beeinflussen.
      */
      
    }

#endif // Server


#ifdef IS_CLIENT
  mpz_t Obergrenze_D;
  mpz_init(Obergrenze_D);
    
#ifdef USE_NETWORK
 ClientMainLoop: // Hauptschleife: Polynomintervall holen und diese durchsieben
  
  {
    unix_io_client communication_stream(communication_name, server_port);
    cout << "Hole dynamische Faktoren..." << endl;
    static streampos dynfac_fpos = 0;
    communication_stream << "DynamicFactors?_ab_fpos " << dynfac_fpos << endl;
    communication_stream >> dynfac_fpos; // nchste Startposition holen
    TDynamicFactorRelation relation;
    relation.fpos=0;
    while (true) // dynamische Relationen holen
      {
	communication_stream >> relation.factor;
	cout << relation.factor << "," << flush;
	if (relation.factor<=0) break;
	relation.SQRT_kN=SQRT_kN_mod_Primzahl(relation.factor);
	Dynamic_Factor_Relations.insert(relation);
      }
    cout << "geholt..." << endl;
  }
  {
    unix_io_client communication_stream(communication_name, server_port);
    cout << "Hole Polynomintervall..." << endl;
    communication_stream << "Polynom? 1000" << endl;
    
    Polynom.Daten_laden(communication_stream);
    communication_stream >> Obergrenze_D;
  }
#else // kein Netzwerk
  Polynom.Daten_laden(communication_stream);
  communication_stream >> Obergrenze_D;
#endif // if Netzwerk else keinNetzwerk
#endif // Client
  
  /* Quadratischen Reste mit Primzahlen aus der Primzahlbasis aussieben */
  do /* solange Relationen aussieben, bis mehr Relationen gefunden sind
	als Primzahlen in der Primzahlbasis sind (oder eine Faktorisierung
	gefunden ist). */
    {
#ifndef IS_CLIENT
      // fr Recovery:
      streampos fpos = Recovery_to_file.tellp();
      Polynom.Daten_sichern(Recovery_to_file); // aktuelles Polynom abspeichern
      Recovery_to_file.seekp(fpos);
#endif
      
      Sieboffset=-M;
      alle_Deltas_neu_berechnen();
      while (Sieboffset<M)
	{
	  Sieb_initialisieren();
	  schnelles_aussieben(); // mit Primzahlen der statischen Faktorbasis
	  aussieben_mit_Quadraten(); // (grere) Quadrate von statischen Faktoren sieben
	  aussieben_mit_Dynamic_Factoren(); // ermglicht weitere Relationen...
	  schnelles_auslesen_des_Siebs();	  
	  Sieboffset+=Siebgroesse;
	}
      Statusmeldung(); // Ausgabe von Informationen
      
      Polynom.naechstes_Polynom_ermitteln();
      
      // collecting_phase_beendet= ( gefundene_Relationen > 0.6*Primbasis_size );
      /* Durch das Flag "collecting_phase_beendet" ist es mglich, das Aufsammeln weiterer dynamischer
	 oder special-Faktoren zu unterbinden. Dies knnte ratsam sein, wenn der Haupt- oder auch
	 der Hintergrundspeicher knapp wird oder das Programm schon solche Fortschritte im
	 Sammeln der eigentlichen Relationen macht, da dynamische oder special-Faktoren hinderlich
	 wirken.
	 Da aber eine Speicherkanppheit bei UNIX nicht sinnvoll ermittelt werden kann (andere Prozesse
	 knnen jederzeit Speicher anfordern/freigeben), empfiehlt es sich eher, das Programm
	 bei Speicherknappheit abzubrechen, die dynamic_relations-Datei zu krzen/zu lschen und
	 das Programm anschlieend neu zu starten. Auf diese Weise kann man die Speichernutzung
	 des Programms wesentlich flexibler beeinflussen.
      */
    }
#ifndef IS_CLIENT
  while (true); // oder pessimistisch: (gefundene_Relationen<=Primbasis_size);
#else
  while (Polynom < Obergrenze_D);
#ifdef USE_NETWORK
  goto ClientMainLoop; // Hauptschleife: neues Polynomintervall holen und fortsetzen
#endif
  mpz_clear (Obergrenze_D);
#endif
  
  delete [] Sieb; /* Sieb lschen, d.h. Speicher wieder freigeben */ 
  
  cout << endl << "Fertig..." << endl;
  return 0;
  
}
